@@ -1,9 +1,9 @@
 /*
- *  $Id: checklist.c,v 1.114 2008/03/16 12:55:06 tom Exp $
+ *  $Id: checklist.c,v 1.131 2011/10/15 12:43:07 tom Exp $
  *
  *  checklist.c -- implements the checklist box
  *
- *  Copyright 2000-2007,2008	Thomas E. Dickey
+ *  Copyright 2000-2010,2011	Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License, version 2.1
@@ -49,13 +49,17 @@ print_arrows(WINDOW *win,
 	     int item_no,
 	     int list_height)
 {
-    dlg_draw_arrows2(win, scrollamt,
-		     scrollamt + choice < item_no,
-		     box_x + check_x + ARROWS_COL,
-		     box_y,
-		     box_y + list_height + 1,
-		     menubox_attr,
-		     menubox_border_attr);
+    dlg_draw_scrollbar(win,
+		       (long) (scrollamt),
+		       (long) (scrollamt),
+		       (long) (scrollamt + choice),
+		       (long) (item_no),
+		       box_x + check_x,
+		       box_x + list_width,
+		       box_y,
+		       box_y + list_height + 1,
+		       menubox_border2_attr,
+		       menubox_border_attr);
 }
 
 /*
@@ -69,7 +73,7 @@ print_item(WINDOW *win,
 	   int choice,
 	   int selected)
 {
-    chtype save = getattrs(win);
+    chtype save = dlg_get_attrs(win);
     int i;
     chtype attr = A_NORMAL;
     const int *cols;
@@ -145,6 +149,7 @@ dlg_checklist(const char *title,
 {
     /* *INDENT-OFF* */
     static DLG_KEYS_BINDING binding[] = {
+	HELPKEY_BINDINGS,
 	ENTERKEY_BINDINGS,
 	DLG_KEYS_DATA( DLGK_FIELD_NEXT, KEY_RIGHT ),
 	DLG_KEYS_DATA( DLGK_FIELD_NEXT, TAB ),
@@ -185,10 +190,32 @@ dlg_checklist(const char *title,
     WINDOW *dialog, *list;
     char *prompt = dlg_strclone(cprompt);
     const char **buttons = dlg_ok_labels();
+    const char *widget_name;
 
     dlg_does_output();
     dlg_tab_correct_str(prompt);
 
+    /*
+     * If this is a radiobutton list, ensure that no more than one item is
+     * selected initially.  Allow none to be selected, since some users may
+     * wish to provide this flavor.
+     */
+    if (flag == FLAG_RADIO) {
+	bool first = TRUE;
+
+	for (i = 0; i < item_no; i++) {
+	    if (items[i].state) {
+		if (first) {
+		    first = FALSE;
+		} else {
+		    items[i].state = 0;
+		}
+	    }
+	}
+	widget_name = "radiolist";
+    } else {
+	widget_name = "checklist";
+    }
 #ifdef KEY_RESIZE
   retry:
 #endif
@@ -202,13 +229,14 @@ dlg_checklist(const char *title,
     } else {
 	dlg_auto_size(title, prompt, &height, &width, MIN_HIGH + use_height, 26);
     }
+    dlg_button_layout(buttons, &width);
     dlg_print_size(height, width);
     dlg_ctl_size(height, width);
 
     /* we need at least two states */
     if (states == 0 || strlen(states) < 2)
 	states = " *";
-    num_states = strlen(states);
+    num_states = (int) strlen(states);
 
     checkflag = flag;
 
@@ -216,13 +244,13 @@ dlg_checklist(const char *title,
     y = dlg_box_y_ordinate(height);
 
     dialog = dlg_new_window(height, width, y, x);
-    dlg_register_window(dialog, "checklist", binding);
-    dlg_register_buttons(dialog, "checklist", buttons);
+    dlg_register_window(dialog, widget_name, binding);
+    dlg_register_buttons(dialog, widget_name, buttons);
 
     dlg_mouse_setbase(x, y);
 
-    dlg_draw_box(dialog, 0, 0, height, width, dialog_attr, border_attr);
-    dlg_draw_bottom_box(dialog);
+    dlg_draw_box2(dialog, 0, 0, height, width, dialog_attr, border_attr, border2_attr);
+    dlg_draw_bottom_box2(dialog, border_attr, border2_attr, dialog_attr);
     dlg_draw_title(dialog, title);
 
     wattrset(dialog, dialog_attr);
@@ -252,7 +280,7 @@ dlg_checklist(const char *title,
     dlg_draw_box(dialog, box_y, box_x,
 		 use_height + 2 * MARGIN,
 		 list_width + 2 * MARGIN,
-		 menubox_border_attr, menubox_attr);
+		 menubox_border_attr, menubox_border2_attr);
 
     text_width = 0;
     name_width = 0;
@@ -286,11 +314,12 @@ dlg_checklist(const char *title,
 	choice = max_choice - 1;
     }
     /* Print the list */
-    for (i = 0; i < max_choice; i++)
+    for (i = 0; i < max_choice; i++) {
 	print_item(list,
 		   &items[i + scrollamt],
 		   states,
 		   i, i == choice);
+    }
     (void) wnoutrefresh(list);
 
     /* register the new window, along with its borders */
@@ -303,6 +332,7 @@ dlg_checklist(const char *title,
 
     dlg_draw_buttons(dialog, height - 2, 0, buttons, button, FALSE, width);
 
+    dlg_trace_win(dialog);
     while (result == DLG_EXIT_UNKNOWN) {
 	if (button < 0)		/* --visit-items */
 	    wmove(dialog, box_y + choice + 1, box_x + check_x + 2);
@@ -547,6 +577,9 @@ dlg_checklist(const char *title,
 			       states,
 			       choice, TRUE);
 		    (void) wnoutrefresh(list);
+		    print_arrows(dialog,
+				 box_x, box_y,
+				 scrollamt, max_choice, item_no, use_height);
 		    (void) wmove(dialog, cur_y, cur_x);
 		    wrefresh(dialog);
 		}
@@ -557,7 +590,7 @@ dlg_checklist(const char *title,
 	if (fkey) {
 	    switch (key) {
 	    case DLGK_ENTER:
-		result = dlg_ok_buttoncode(button);
+		result = dlg_enter_buttoncode(button);
 		break;
 	    case DLGK_FIELD_PREV:
 		button = dlg_prev_button(buttons, button);
@@ -624,16 +657,18 @@ dialog_checklist(const char *title,
     bool show_status = FALSE;
     int current = 0;
 
-    listitems = dlg_calloc(DIALOG_LISTITEM, item_no + 1);
+    listitems = dlg_calloc(DIALOG_LISTITEM, (size_t) item_no + 1);
     assert_ptr(listitems, "dialog_checklist");
 
     for (i = 0; i < item_no; ++i) {
 	listitems[i].name = ItemName(i);
 	listitems[i].text = ItemText(i);
-	listitems[i].help = (dialog_vars.item_help) ? ItemHelp(i) :
-	    dlg_strempty();
+	listitems[i].help = ((dialog_vars.item_help)
+			     ? ItemHelp(i)
+			     : dlg_strempty());
 	listitems[i].state = !dlg_strcmp(ItemStatus(i), "on");
     }
+    dlg_align_columns(&listitems[0].text, (int) sizeof(DIALOG_LISTITEM), item_no);
 
     result = dlg_checklist(title,
 			   cprompt,
@@ -657,25 +692,25 @@ dialog_checklist(const char *title,
 	if (USE_ITEM_HELP(listitems[current].help)) {
 	    if (show_status) {
 		if (separate_output) {
-		    dlg_add_result(listitems[current].help);
-		    dlg_add_result("\n");
+		    dlg_add_string(listitems[current].help);
+		    dlg_add_separator();
 		} else {
 		    dlg_add_quoted(listitems[current].help);
 		}
 	    } else {
-		dlg_add_result(listitems[current].help);
+		dlg_add_string(listitems[current].help);
 	    }
 	    result = DLG_EXIT_ITEM_HELP;
 	} else {
 	    if (show_status) {
 		if (separate_output) {
-		    dlg_add_result(listitems[current].name);
-		    dlg_add_result("\n");
+		    dlg_add_string(listitems[current].name);
+		    dlg_add_separator();
 		} else {
 		    dlg_add_quoted(listitems[current].name);
 		}
 	    } else {
-		dlg_add_result(listitems[current].name);
+		dlg_add_string(listitems[current].name);
 	    }
 	}
 	break;
@@ -685,21 +720,18 @@ dialog_checklist(const char *title,
 	for (i = 0; i < item_no; i++) {
 	    if (listitems[i].state) {
 		if (separate_output) {
-		    dlg_add_result(listitems[i].name);
-		    dlg_add_result("\n");
+		    dlg_add_string(listitems[i].name);
+		    dlg_add_separator();
 		} else {
-		    if (dialog_vars.input_result && *(dialog_vars.input_result))
-			dlg_add_result(" ");
-		    if (flag == FLAG_CHECK) {
-			dlg_add_quoted(listitems[i].name);
-		    } else {
-			dlg_add_result(listitems[i].name);
-		    }
+		    if (dlg_need_separator())
+			dlg_add_separator();
+		    dlg_add_string(listitems[i].name);
 		}
 	    }
 	}
     }
 
+    dlg_free_columns(&listitems[0].text, (int) sizeof(DIALOG_LISTITEM), item_no);
     free(listitems);
     return result;
 }
