@@ -1,9 +1,9 @@
 /*
- *  $Id: editbox.c,v 1.46 2008/06/21 12:07:54 tom Exp $
+ *  $Id: editbox.c,v 1.58 2011/10/15 12:57:37 tom Exp $
  *
  *  editbox.c -- implements the edit box
  *
- *  Copyright 2007,2008 Thomas E. Dickey
+ *  Copyright 2007-2010,2011 Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License, version 2.1
@@ -37,9 +37,9 @@ static void
 grow_list(char ***list, int *have, int want)
 {
     if (want > *have) {
-	unsigned last = *have;
-	unsigned need = (want | 31) + 3;
-	*have = need;
+	size_t last = (size_t) *have;
+	size_t need = (size_t) (want | 31) + 3;
+	*have = (int) need;
 	(*list) = dlg_realloc(char *, need, *list);
 	if ((*list) == 0) {
 	    fail_list();
@@ -58,7 +58,7 @@ load_list(const char *file, char ***list, int *rows)
     struct stat sb;
     unsigned n, pass;
     unsigned need;
-    unsigned size;
+    size_t size;
 
     *list = 0;
     *rows = 0;
@@ -67,7 +67,7 @@ load_list(const char *file, char ***list, int *rows)
 	(sb.st_mode & S_IFMT) != S_IFREG)
 	dlg_exiterr("Not a file: %s", file);
 
-    size = sb.st_size;
+    size = (size_t) sb.st_size;
     if ((blob = dlg_malloc(char, size + 1)) == 0)
 	  fail_list();
     blob[size] = '\0';
@@ -103,7 +103,7 @@ load_list(const char *file, char ***list, int *rows)
 		(*list)[need] = 0;
 	    }
 	} else {
-	    grow_list(list, rows, need + 1);
+	    grow_list(list, rows, (int) need + 1);
 	}
     }
     free(blob);
@@ -247,7 +247,7 @@ col_to_chr_offset(const char *text, int col)
     bool found = FALSE;
     int result = 0;
     unsigned n;
-    unsigned len = dlg_count_wchars(text);
+    unsigned len = (unsigned) dlg_count_wchars(text);
 
     for (n = 0; n < len; ++n) {
 	if (cols[n] <= col && cols[n + 1] > col) {
@@ -270,6 +270,36 @@ col_to_chr_offset(const char *text, int col)
 
 #define UPDATE_COL(input) col_offset = dlg_edit_offset(input, chr_offset, box_width)
 
+static int
+widest_line(char **list)
+{
+    int result = MAX_LEN;
+    char *value;
+
+    if (list != 0) {
+	while ((value = *list++) != 0) {
+	    int check = (int) strlen(value);
+	    if (check > result)
+		result = check;
+	}
+    }
+    return result;
+}
+
+#define NAVIGATE_BINDINGS \
+	DLG_KEYS_DATA( DLGK_GRID_DOWN,	KEY_DOWN ), \
+	DLG_KEYS_DATA( DLGK_GRID_RIGHT,	KEY_RIGHT ), \
+	DLG_KEYS_DATA( DLGK_GRID_LEFT,	KEY_LEFT ), \
+	DLG_KEYS_DATA( DLGK_GRID_UP,	KEY_UP ), \
+	DLG_KEYS_DATA( DLGK_FIELD_NEXT,	TAB ), \
+	DLG_KEYS_DATA( DLGK_FIELD_PREV,	KEY_BTAB ), \
+	DLG_KEYS_DATA( DLGK_PAGE_FIRST,	KEY_HOME ), \
+	DLG_KEYS_DATA( DLGK_PAGE_LAST,	KEY_END ), \
+	DLG_KEYS_DATA( DLGK_PAGE_LAST,	KEY_LL ), \
+	DLG_KEYS_DATA( DLGK_PAGE_NEXT,	KEY_NPAGE ), \
+	DLG_KEYS_DATA( DLGK_PAGE_NEXT,	DLGK_MOUSE(KEY_NPAGE) ), \
+	DLG_KEYS_DATA( DLGK_PAGE_PREV,	KEY_PPAGE ), \
+	DLG_KEYS_DATA( DLGK_PAGE_PREV,	DLGK_MOUSE(KEY_PPAGE) )
 /*
  * Display a dialog box for editing a copy of a file
  */
@@ -282,21 +312,16 @@ dlg_editbox(const char *title,
 {
     /* *INDENT-OFF* */
     static DLG_KEYS_BINDING binding[] = {
-	INPUTSTR_BINDINGS,
+	HELPKEY_BINDINGS,
 	ENTERKEY_BINDINGS,
-	DLG_KEYS_DATA( DLGK_GRID_DOWN,	KEY_DOWN ),
-	DLG_KEYS_DATA( DLGK_GRID_RIGHT,	KEY_RIGHT ),
-	DLG_KEYS_DATA( DLGK_GRID_LEFT,	KEY_LEFT ),
-	DLG_KEYS_DATA( DLGK_GRID_UP,	KEY_UP ),
-	DLG_KEYS_DATA( DLGK_FIELD_NEXT,	TAB ),
-	DLG_KEYS_DATA( DLGK_FIELD_PREV,	KEY_BTAB ),
-	DLG_KEYS_DATA( DLGK_PAGE_FIRST,	KEY_HOME ),
-	DLG_KEYS_DATA( DLGK_PAGE_LAST,	KEY_END ),
-	DLG_KEYS_DATA( DLGK_PAGE_LAST,	KEY_LL ),
-	DLG_KEYS_DATA( DLGK_PAGE_NEXT,	KEY_NPAGE ),
-	DLG_KEYS_DATA( DLGK_PAGE_NEXT,	DLGK_MOUSE(KEY_NPAGE) ),
-	DLG_KEYS_DATA( DLGK_PAGE_PREV,	KEY_PPAGE ),
-	DLG_KEYS_DATA( DLGK_PAGE_PREV,	DLGK_MOUSE(KEY_PPAGE) ),
+	NAVIGATE_BINDINGS,
+	END_KEYS_BINDING
+    };
+    static DLG_KEYS_BINDING binding2[] = {
+	INPUTSTR_BINDINGS,
+	HELPKEY_BINDINGS,
+	ENTERKEY_BINDINGS,
+	NAVIGATE_BINDINGS,
 	END_KEYS_BINDING
     };
     /* *INDENT-ON* */
@@ -316,20 +341,22 @@ dlg_editbox(const char *title,
     int listsize = size_list(*list);
     int result = DLG_EXIT_UNKNOWN;
     int state;
-    int max_len = dlg_max_input(MAX_LEN);
+    size_t max_len = (size_t) dlg_max_input(widest_line(*list));
     char *input, *buffer;
     bool show_all, show_one, was_mouse;
+    bool first_trace = TRUE;
     WINDOW *dialog;
     WINDOW *editing;
     DIALOG_VARS save_vars;
     const char **buttons = dlg_ok_labels();
+    int mincols = (3 * COLS / 4);
 
     dlg_save_vars(&save_vars);
     dialog_vars.separate_output = TRUE;
 
     dlg_does_output();
 
-    buffer = dlg_malloc(char, max_len);
+    buffer = dlg_malloc(char, max_len + 1);
     assert_ptr(buffer, "dlg_editbox");
 
     thisrow = base_row = lastrow = 0;
@@ -341,7 +368,8 @@ dlg_editbox(const char *title,
     state = dialog_vars.defaultno ? dlg_defaultno_button() : sTEXT;
     key = fkey = 0;
 
-    dlg_auto_size(title, "", &height, &width, 3 * LINES / 4, 3 * COLS / 4);
+    dlg_button_layout(buttons, &mincols);
+    dlg_auto_size(title, "", &height, &width, 3 * LINES / 4, mincols);
     dlg_print_size(height, width);
     dlg_ctl_size(height, width);
 
@@ -354,8 +382,8 @@ dlg_editbox(const char *title,
 
     dlg_mouse_setbase(x, y);
 
-    dlg_draw_box(dialog, 0, 0, height, width, dialog_attr, border_attr);
-    dlg_draw_bottom_box(dialog);
+    dlg_draw_box2(dialog, 0, 0, height, width, dialog_attr, border_attr, border2_attr);
+    dlg_draw_bottom_box2(dialog, border_attr, border2_attr, dialog_attr);
     dlg_draw_title(dialog, title);
 
     wattrset(dialog, dialog_attr);
@@ -371,7 +399,7 @@ dlg_editbox(const char *title,
 		 box_x,
 		 box_height,
 		 box_width,
-		 border_attr, dialog_attr);
+		 border_attr, border2_attr);
     dlg_mouse_mkbigregion(box_y + MARGIN,
 			  box_x + MARGIN,
 			  box_height - (2 * MARGIN),
@@ -382,7 +410,7 @@ dlg_editbox(const char *title,
 			     box_width - (2 * MARGIN),
 			     getbegy(dialog) + box_y + 1,
 			     getbegx(dialog) + box_x + 1);
-    dlg_register_window(editing, "editbox", binding);
+    dlg_register_window(editing, "editbox2", binding2);
 
     show_all = TRUE;
     show_one = FALSE;
@@ -408,14 +436,17 @@ dlg_editbox(const char *title,
 	    display_one(editing, THIS_ROW,
 			thisrow, thisrow, base_row, chr_offset);
 	    getyx(editing, y, x);
-	    dlg_draw_arrows2(dialog,
-			     base_row != 0,
-			     base_row + pagesize < listsize,
-			     box_x + ARROWS_COL,
-			     box_y + 0,
-			     box_y + getmaxy(editing) + 1,
-			     dialog_attr,
-			     border_attr);
+	    dlg_draw_scrollbar(dialog,
+			       base_row,
+			       base_row,
+			       base_row + pagesize,
+			       listsize,
+			       box_x,
+			       box_x + getmaxx(editing),
+			       box_y + 0,
+			       box_y + getmaxy(editing) + 1,
+			       border2_attr,
+			       border_attr);
 	    wmove(editing, y, x);
 	    show_one = FALSE;
 	}
@@ -446,7 +477,19 @@ dlg_editbox(const char *title,
 	    }
 	}
 
+	if (first_trace) {
+	    first_trace = FALSE;
+	    dlg_trace_win(dialog);
+	}
+
 	key = dlg_mouse_wgetch((state == sTEXT) ? editing : dialog, &fkey);
+	if (key == ERR) {
+	    result = DLG_EXIT_ERROR;
+	    break;
+	} else if (key == ESC) {
+	    result = DLG_EXIT_ESC;
+	    break;
+	}
 	if (state != sTEXT) {
 	    if (dlg_result_key(key, fkey, &result))
 		break;
@@ -520,7 +563,8 @@ dlg_editbox(const char *title,
 			if (thisrow == 0) {
 			    beep();
 			} else {
-			    int len = strlen(THIS_ROW) + strlen(PREV_ROW) + 1;
+			    size_t len = (strlen(THIS_ROW) +
+					  strlen(PREV_ROW) + 1);
 			    char *tmp = dlg_malloc(char, len);
 
 			    assert_ptr(tmp, "dlg_editbox");
@@ -564,7 +608,7 @@ dlg_editbox(const char *title,
 		    continue;
 		}
 	    }
-	    strcpy(buffer, input);
+	    strncpy(buffer, input, max_len - 1)[max_len - 1] = '\0';
 	    edit = dlg_edit_string(buffer, &chr_offset, key, fkey, FALSE);
 
 	    if (edit) {
