@@ -1,9 +1,9 @@
 /*
- *  $Id: formbox.c,v 1.63 2008/03/16 18:37:12 Vladimir.Mezentsev Exp $
+ *  $Id: formbox.c,v 1.79 2011/10/15 12:45:11 tom Exp $
  *
  *  formbox.c -- implements the form (i.e, some pairs label/editbox)
  *
- *  Copyright 2003-2007,2008	Thomas E. Dickey
+ *  Copyright 2003-2010,2011	Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License, version 2.1
@@ -188,6 +188,34 @@ form_limit(DIALOG_FORMITEM item[])
     return limit;
 }
 
+static int
+is_first_field(DIALOG_FORMITEM item[], int choice)
+{
+    int count = 0;
+    while (choice >= 0) {
+	if (item[choice].text_flen > 0) {
+	    ++count;
+	}
+	--choice;
+    }
+
+    return (count == 1);
+}
+
+static int
+is_last_field(DIALOG_FORMITEM item[], int choice, int item_no)
+{
+    int count = 0;
+    while (choice < item_no) {
+	if (item[choice].text_flen > 0) {
+	    ++count;
+	}
+	++choice;
+    }
+
+    return (count == 1);
+}
+
 /*
  * Tab to the next field.
  */
@@ -351,7 +379,7 @@ make_FORM_ELTs(DIALOG_FORMITEM * item,
 	    int max_len = dlg_max_input(MAX(item[i].text_ilen + 1, MAX_LEN));
 	    char *old_text = item[i].text;
 
-	    item[i].text = dlg_malloc(char, max_len + 1);
+	    item[i].text = dlg_malloc(char, (size_t) max_len + 1);
 	    assert_ptr(item[i].text, "make_FORM_ELTs");
 
 	    sprintf(item[i].text, "%.*s", item[i].text_ilen, old_text);
@@ -407,6 +435,17 @@ prev_valid_buttonindex(int state, int extra, bool non_editable)
     return state;
 }
 
+#define NAVIGATE_BINDINGS \
+	DLG_KEYS_DATA( DLGK_FIELD_NEXT, TAB ), \
+	DLG_KEYS_DATA( DLGK_FIELD_PREV, KEY_BTAB ), \
+	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  CHR_NEXT ), \
+	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  KEY_DOWN ), \
+	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  KEY_NEXT ), \
+	DLG_KEYS_DATA( DLGK_ITEM_PREV,  CHR_PREVIOUS ), \
+	DLG_KEYS_DATA( DLGK_ITEM_PREV,  KEY_PREVIOUS ), \
+	DLG_KEYS_DATA( DLGK_ITEM_PREV,  KEY_UP ), \
+	DLG_KEYS_DATA( DLGK_PAGE_NEXT,  KEY_NPAGE ), \
+	DLG_KEYS_DATA( DLGK_PAGE_PREV,  KEY_PPAGE )
 /*
  * Display a form for fulfill a number of fields
  */
@@ -422,18 +461,16 @@ dlg_form(const char *title,
 {
     /* *INDENT-OFF* */
     static DLG_KEYS_BINDING binding[] = {
-	INPUTSTR_BINDINGS,
+	HELPKEY_BINDINGS,
 	ENTERKEY_BINDINGS,
-	DLG_KEYS_DATA( DLGK_FIELD_NEXT, TAB ),
-	DLG_KEYS_DATA( DLGK_FIELD_PREV, KEY_BTAB ),
-	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  CHR_NEXT ),
-	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  KEY_DOWN ),
-	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  KEY_NEXT ),
-	DLG_KEYS_DATA( DLGK_ITEM_PREV,  CHR_PREVIOUS ),
-	DLG_KEYS_DATA( DLGK_ITEM_PREV,  KEY_PREVIOUS ),
-	DLG_KEYS_DATA( DLGK_ITEM_PREV,  KEY_UP ),
-	DLG_KEYS_DATA( DLGK_PAGE_NEXT,  KEY_NPAGE ),
-	DLG_KEYS_DATA( DLGK_PAGE_PREV,  KEY_PPAGE ),
+	NAVIGATE_BINDINGS,
+	END_KEYS_BINDING
+    };
+    static DLG_KEYS_BINDING binding2[] = {
+	INPUTSTR_BINDINGS,
+	HELPKEY_BINDINGS,
+	ENTERKEY_BINDINGS,
+	NAVIGATE_BINDINGS,
 	END_KEYS_BINDING
     };
     /* *INDENT-ON* */
@@ -445,6 +482,7 @@ dlg_form(const char *title,
 
     int form_width;
     int first = TRUE;
+    int first_trace = TRUE;
     int chr_offset = 0;
     int state = dialog_vars.defaultno ? dlg_defaultno_button() : sTEXT;
     int x, y, cur_x, cur_y, box_x, box_y;
@@ -467,6 +505,7 @@ dlg_form(const char *title,
     DIALOG_FORMITEM *current;
 
     make_FORM_ELTs(items, item_no, &min_height, &min_width);
+    dlg_button_layout(buttons, &min_width);
     dlg_does_output();
     dlg_tab_correct_str(prompt);
 
@@ -504,8 +543,8 @@ dlg_form(const char *title,
 
     dlg_mouse_setbase(x, y);
 
-    dlg_draw_box(dialog, 0, 0, height, width, dialog_attr, border_attr);
-    dlg_draw_bottom_box(dialog);
+    dlg_draw_box2(dialog, 0, 0, height, width, dialog_attr, border_attr, border2_attr);
+    dlg_draw_bottom_box2(dialog, border_attr, border2_attr, dialog_attr);
     dlg_draw_title(dialog, title);
 
     wattrset(dialog, dialog_attr);
@@ -513,16 +552,18 @@ dlg_form(const char *title,
 
     form_width = width - 6;
     getyx(dialog, cur_y, cur_x);
+    (void) cur_x;
     box_y = cur_y + 1;
     box_x = (width - form_width) / 2 - 1;
 
     /* create new window for the form */
     form = dlg_sub_window(dialog, form_height, form_width, y + box_y + 1,
 			  x + box_x + 1);
+    dlg_register_window(form, "formfield", binding2);
 
     /* draw a box around the form items */
     dlg_draw_box(dialog, box_y, box_x, form_height + 2, form_width + 2,
-		 menubox_border_attr, menubox_attr);
+		 menubox_border_attr, menubox_border2_attr);
 
     /* register the new window, along with its borders */
     dlg_mouse_mkbigregion(getbegy(form) - getbegy(dialog),
@@ -544,11 +585,17 @@ dlg_form(const char *title,
 
 	if (scroll_changed) {
 	    print_form(form, items, item_no, scrollamt, choice);
-	    dlg_draw_arrows(dialog, scrollamt,
-			    scrollamt + form_height < item_no,
-			    box_x + 1,
-			    box_y,
-			    box_y + form_height + 1);
+	    dlg_draw_scrollbar(dialog,
+			       scrollamt,
+			       scrollamt,
+			       scrollamt + form_height + 1,
+			       min_height,
+			       box_x + 1,
+			       box_x + form_width,
+			       box_y,
+			       box_y + form_height + 1,
+			       menubox_border2_attr,
+			       menubox_border_attr);
 	    scroll_changed = FALSE;
 	}
 
@@ -562,6 +609,11 @@ dlg_form(const char *title,
 	    show_buttons = FALSE;
 	}
 
+	if (first_trace) {
+	    first_trace = FALSE;
+	    dlg_trace_win(dialog);
+	}
+
 	if (field_changed || state == sTEXT) {
 	    if (field_changed)
 		chr_offset = 0;
@@ -577,7 +629,7 @@ dlg_form(const char *title,
 	    field_changed = FALSE;
 	}
 
-	key = dlg_mouse_wgetch(dialog, &fkey);
+	key = dlg_mouse_wgetch((state == sTEXT) ? form : dialog, &fkey);
 	if (dlg_result_key(key, fkey, &result))
 	    break;
 
@@ -618,7 +670,7 @@ dlg_form(const char *title,
 
 	    case DLGK_ENTER:
 		dlg_del_window(dialog);
-		result = (state >= 0) ? dlg_ok_buttoncode(state) : DLG_EXIT_OK;
+		result = (state >= 0) ? dlg_enter_buttoncode(state) : DLG_EXIT_OK;
 		continue;
 
 	    case DLGK_GRID_LEFT:
@@ -636,6 +688,25 @@ dlg_form(const char *title,
 		    continue;
 		}
 
+	    case DLGK_FORM_PREV:
+		if (state == sTEXT && !is_first_field(items, choice)) {
+		    do_tab = TRUE;
+		    move_by = -1;
+		    break;
+		} else {
+		    int old_state = state;
+		    state = prev_valid_buttonindex(state, sTEXT, non_editable);
+		    show_buttons = TRUE;
+		    if (old_state >= 0 && state == sTEXT) {
+			new_choice = item_no - 1;
+			if (choice != new_choice) {
+			    print_item(form, items + choice, scrollamt, FALSE);
+			    choice = new_choice;
+			}
+		    }
+		    continue;
+		}
+
 	    case DLGK_FIELD_PREV:
 		state = prev_valid_buttonindex(state, sTEXT, non_editable);
 		show_buttons = TRUE;
@@ -662,6 +733,21 @@ dlg_form(const char *title,
 		    continue;
 		}
 
+	    case DLGK_FORM_NEXT:
+		if (state == sTEXT && !is_last_field(items, choice, item_no)) {
+		    do_tab = TRUE;
+		    move_by = 1;
+		    break;
+		} else {
+		    state = next_valid_buttonindex(state, sTEXT, non_editable);
+		    show_buttons = TRUE;
+		    if (state == sTEXT && choice) {
+			print_item(form, items + choice, scrollamt, FALSE);
+			choice = 0;
+		    }
+		    continue;
+		}
+
 #ifdef KEY_RESIZE
 	    case KEY_RESIZE:
 		/* reset data */
@@ -818,25 +904,30 @@ dialog_form(const char *title,
     int choice;
     int i;
     DIALOG_FORMITEM *listitems;
+    DIALOG_VARS save_vars;
     bool show_status = FALSE;
 
-    listitems = dlg_calloc(DIALOG_FORMITEM, item_no + 1);
+    dlg_save_vars(&save_vars);
+    dialog_vars.separate_output = TRUE;
+
+    listitems = dlg_calloc(DIALOG_FORMITEM, (size_t) item_no + 1);
     assert_ptr(listitems, "dialog_form");
 
     for (i = 0; i < item_no; ++i) {
 	listitems[i].type = dialog_vars.formitem_type;
 	listitems[i].name = ItemName(i);
-	listitems[i].name_len = strlen(ItemName(i));
+	listitems[i].name_len = (int) strlen(ItemName(i));
 	listitems[i].name_y = dlg_ordinate(ItemNameY(i));
 	listitems[i].name_x = dlg_ordinate(ItemNameX(i));
 	listitems[i].text = ItemText(i);
-	listitems[i].text_len = strlen(ItemText(i));
+	listitems[i].text_len = (int) strlen(ItemText(i));
 	listitems[i].text_y = dlg_ordinate(ItemTextY(i));
 	listitems[i].text_x = dlg_ordinate(ItemTextX(i));
 	listitems[i].text_flen = atoi(ItemTextFLen(i));
 	listitems[i].text_ilen = atoi(ItemTextILen(i));
-	listitems[i].help = (dialog_vars.item_help) ? ItemHelp(i) :
-	    dlg_strempty();
+	listitems[i].help = ((dialog_vars.item_help)
+			     ? ItemHelp(i)
+			     : dlg_strempty());
     }
 
     result = dlg_form(title,
@@ -857,25 +948,26 @@ dialog_form(const char *title,
 	dlg_add_result("HELP ");
 	show_status = dialog_vars.help_status;
 	if (USE_ITEM_HELP(listitems[choice].help)) {
-	    dlg_add_result(listitems[choice].help);
+	    dlg_add_string(listitems[choice].help);
 	    result = DLG_EXIT_ITEM_HELP;
 	} else {
-	    dlg_add_result(listitems[choice].name);
+	    dlg_add_string(listitems[choice].name);
 	}
 	if (show_status)
-	    dlg_add_result("\n");
+	    dlg_add_separator();
 	break;
     }
     if (show_status) {
 	for (i = 0; i < item_no; i++) {
 	    if (listitems[i].text_flen > 0) {
-		dlg_add_result(listitems[i].text);
-		dlg_add_result("\n");
+		dlg_add_string(listitems[i].text);
+		dlg_add_separator();
 	    }
 	}
     }
 
     dlg_free_formitems(listitems);
+    dlg_restore_vars(&save_vars);
 
     return result;
 }
