@@ -1,9 +1,9 @@
 /*
- * $Id: fselect.c,v 1.69 2007/02/22 23:56:00 tom Exp $
+ *  $Id: fselect.c,v 1.83 2011/10/16 19:30:12 tom Exp $
  *
- * fselect.c -- implements the file-selector box
+ *  fselect.c -- implements the file-selector box
  *
- * Copyright 2000-2006,2007 Thomas E. Dickey
+ *  Copyright 2000-2010,2011	Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License, version 2.1
@@ -142,7 +142,7 @@ add_to_list(LIST * list, char *text)
 {
     unsigned need;
 
-    need = list->length + 1;
+    need = (unsigned) (list->length + 1);
     if (need + 1 > list->allocd) {
 	list->allocd = 2 * (need + 1);
 	if (list->data == 0) {
@@ -240,11 +240,17 @@ display_list(LIST * list)
 
 	top = y - 1;
 	bottom = y + getmaxy(list->win);
-	dlg_draw_arrows(list->par, list->offset,
-			list->length - list->offset > getmaxy(list->win),
-			x + 1,
-			top,
-			bottom);
+	dlg_draw_scrollbar(list->par,
+			   (long) list->offset,
+			   (long) list->offset,
+			   (long) (list->offset + getmaxy(list->win)),
+			   (long) (list->length),
+			   x + 1,
+			   x + getmaxx(list->win),
+			   top,
+			   bottom,
+			   menubox_border2_attr,
+			   menubox_border_attr);
 
 	(void) wmove(list->win, list->choice - list->offset, 0);
 	(void) wnoutrefresh(list->win);
@@ -264,18 +270,20 @@ fix_arrows(LIST * list)
     int x;
     int y;
     int top;
+    int right;
     int bottom;
 
     if (list->win != 0) {
 	getparyx(list->win, y, x);
 	top = y - 1;
+	right = getmaxx(list->win);
 	bottom = y + getmaxy(list->win);
 
-	mouse_mkbutton(top, x, 6,
+	mouse_mkbutton(top, x, right,
 		       ((list->mousex == MOUSE_D)
 			? KEY_PREVIOUS
 			: KEY_PPAGE));
-	mouse_mkbutton(bottom, x, 6,
+	mouse_mkbutton(bottom, x, right,
 		       ((list->mousex == MOUSE_D)
 			? KEY_NEXT
 			: KEY_NPAGE));
@@ -283,7 +291,7 @@ fix_arrows(LIST * list)
 }
 
 static int
-show_list(char *target, LIST * list, bool keep)
+show_list(char *target, LIST * list, int keep)
 {
     int changed = keep || find_choice(target, list);
     display_list(list);
@@ -295,7 +303,7 @@ show_list(char *target, LIST * list, bool keep)
  * to match.
  */
 static int
-show_both_lists(char *input, LIST * d_list, LIST * f_list, bool keep)
+show_both_lists(char *input, LIST * d_list, LIST * f_list, int keep)
 {
     char *leaf = leaf_of(input);
 
@@ -341,13 +349,13 @@ compar(const void *a, const void *b)
     return strcmp(*(const char *const *) a, *(const char *const *) b);
 }
 
-static int
+static void
 match(char *name, LIST * d_list, LIST * f_list, MATCH * match_list)
 {
     char *test = leaf_of(name);
-    int test_len = strlen(test);
-    char **matches = dlg_malloc(char *, (d_list->length + f_list->length));
-    int data_len = 0;
+    size_t test_len = strlen(test);
+    char **matches = dlg_malloc(char *, (size_t) (d_list->length + f_list->length));
+    size_t data_len = 0;
     int i;
     for (i = 2; i < d_list->length; i++) {
 	if (strncmp(test, d_list->data[i], test_len) == 0) {
@@ -361,8 +369,7 @@ match(char *name, LIST * d_list, LIST * f_list, MATCH * match_list)
     }
     matches = dlg_realloc(char *, data_len, matches);
     match_list->data = matches;
-    match_list->length = data_len;
-    return data_len;
+    match_list->length = (int) data_len;
 }
 
 static void
@@ -377,14 +384,17 @@ complete(char *name, LIST * d_list, LIST * f_list, char **buff_ptr)
 {
     MATCH match_list;
     char *test;
-    int test_len;
-    int i, j;
+    size_t test_len;
+    size_t i;
+    int j;
     char *buff;
+
     match(name, d_list, f_list, &match_list);
     if (match_list.length == 0) {
 	*buff_ptr = NULL;
 	return 0;
     }
+
     test = match_list.data[0];
     test_len = strlen(test);
     buff = dlg_malloc(char, test_len + 2);
@@ -415,12 +425,14 @@ complete(char *name, LIST * d_list, LIST * f_list, char **buff_ptr)
     free_match(&match_list);
     buff[i] = '\0';
     *buff_ptr = buff;
-    return i;
+    return (i != 0);
 }
 
 static bool
-fill_lists(char *current, char *input, LIST * d_list, LIST * f_list, bool keep)
+fill_lists(char *current, char *input, LIST * d_list, LIST * f_list, int keep)
 {
+    bool result = TRUE;
+    bool rescan = FALSE;
     DIR *dp;
     DIRENT *de;
     struct stat sb;
@@ -433,45 +445,60 @@ fill_lists(char *current, char *input, LIST * d_list, LIST * f_list, bool keep)
 	if (current[n] != input[n])
 	    break;
     }
-    if (current[n] == input[n])
-	return FALSE;
-    if (strchr(current + n, '/') == 0
-	&& strchr(input + n, '/') == 0) {
-	return show_both_lists(input, d_list, f_list, keep);
-    }
 
-    strcpy(current, input);
-
-    /* refill the lists */
-    free_list(d_list, TRUE);
-    free_list(f_list, TRUE);
-    strcpy(path, current);
-    if ((leaf = strrchr(path, '/')) != 0) {
-	*++leaf = 0;
+    if (current[n] == input[n]) {
+	result = FALSE;
+	rescan = (n == 0 && d_list->length == 0);
+    } else if (strchr(current + n, '/') == 0
+	       && strchr(input + n, '/') == 0) {
+	result = show_both_lists(input, d_list, f_list, keep);
     } else {
-	strcpy(path, "./");
-	leaf = path + strlen(path);
-    }
-    if ((dp = opendir(path)) != 0) {
-	while ((de = readdir(dp)) != 0) {
-	    strncpy(leaf, de->d_name, NAMLEN(de))[NAMLEN(de)] = 0;
-	    if (stat(path, &sb) == 0) {
-		if ((sb.st_mode & S_IFMT) == S_IFDIR)
-		    add_to_list(d_list, leaf);
-		else if (f_list->win)
-		    add_to_list(f_list, leaf);
-	    }
-	}
-	(void) closedir(dp);
-	/* sort the lists */
-	qsort(d_list->data, d_list->length, sizeof(d_list->data[0]), compar);
-	qsort(f_list->data, f_list->length, sizeof(f_list->data[0]), compar);
+	rescan = TRUE;
     }
 
-    (void) show_both_lists(input, d_list, f_list, FALSE);
-    d_list->offset = d_list->choice;
-    f_list->offset = f_list->choice;
-    return TRUE;
+    if (rescan) {
+
+	strcpy(current, input);
+
+	/* refill the lists */
+	free_list(d_list, TRUE);
+	free_list(f_list, TRUE);
+	strcpy(path, current);
+	if ((leaf = strrchr(path, '/')) != 0) {
+	    *++leaf = 0;
+	} else {
+	    strcpy(path, "./");
+	    leaf = path + strlen(path);
+	}
+	dlg_trace_msg("opendir '%s'\n", path);
+	if ((dp = opendir(path)) != 0) {
+	    while ((de = readdir(dp)) != 0) {
+		strncpy(leaf, de->d_name, NAMLEN(de))[NAMLEN(de)] = 0;
+		if (stat(path, &sb) == 0) {
+		    if ((sb.st_mode & S_IFMT) == S_IFDIR)
+			add_to_list(d_list, leaf);
+		    else if (f_list->win)
+			add_to_list(f_list, leaf);
+		}
+	    }
+	    (void) closedir(dp);
+	    /* sort the lists */
+	    qsort(d_list->data,
+		  (size_t) d_list->length,
+		  sizeof(d_list->data[0]),
+		  compar);
+	    qsort(f_list->data,
+		  (size_t) f_list->length,
+		  sizeof(f_list->data[0]),
+		  compar);
+	}
+
+	(void) show_both_lists(input, d_list, f_list, FALSE);
+	d_list->offset = d_list->choice;
+	f_list->offset = f_list->choice;
+	result = TRUE;
+    }
+    return result;
 }
 
 static bool
@@ -498,25 +525,36 @@ usable_state(int state, LIST * dirs, LIST * files)
 			: ((state == sDIRS) \
 			  ? &d_list \
 			  : 0))
+#define NAVIGATE_BINDINGS \
+	DLG_KEYS_DATA( DLGK_FIELD_NEXT, KEY_RIGHT ), \
+	DLG_KEYS_DATA( DLGK_FIELD_NEXT, TAB ), \
+	DLG_KEYS_DATA( DLGK_FIELD_PREV, KEY_BTAB ), \
+	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  KEY_DOWN ), \
+	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  CHR_NEXT ), \
+	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  KEY_NEXT ), \
+	DLG_KEYS_DATA( DLGK_ITEM_PREV,  CHR_PREVIOUS ), \
+	DLG_KEYS_DATA( DLGK_ITEM_PREV,  KEY_UP ), \
+	DLG_KEYS_DATA( DLGK_PAGE_NEXT,  KEY_NPAGE ), \
+	DLG_KEYS_DATA( DLGK_PAGE_PREV,  KEY_PPAGE )
+
 /*
  * Display a dialog box for entering a filename
  */
 static int
-dlg_fselect(const char *title, const char *path, int height, int width, bool dselect)
+dlg_fselect(const char *title, const char *path, int height, int width, int dselect)
 {
     /* *INDENT-OFF* */
     static DLG_KEYS_BINDING binding[] = {
-	DLG_KEYS_DATA( DLGK_FIELD_PREV, KEY_LEFT ),	/* override inputstr */
-	INPUTSTR_BINDINGS,
+	HELPKEY_BINDINGS,
 	ENTERKEY_BINDINGS,
-	DLG_KEYS_DATA( DLGK_FIELD_NEXT, KEY_RIGHT ),
-	DLG_KEYS_DATA( DLGK_FIELD_NEXT, TAB ),
-	DLG_KEYS_DATA( DLGK_FIELD_PREV, KEY_BTAB ),
-	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  KEY_DOWN),
-	DLG_KEYS_DATA( DLGK_ITEM_NEXT,  KEY_NEXT ),
-	DLG_KEYS_DATA( DLGK_ITEM_PREV,  KEY_UP ),
-	DLG_KEYS_DATA( DLGK_PAGE_NEXT,  KEY_NPAGE ),
-	DLG_KEYS_DATA( DLGK_PAGE_PREV,  KEY_PPAGE ),
+	NAVIGATE_BINDINGS,
+	END_KEYS_BINDING
+    };
+    static DLG_KEYS_BINDING binding2[] = {
+	INPUTSTR_BINDINGS,
+	HELPKEY_BINDINGS,
+	ENTERKEY_BINDINGS,
+	NAVIGATE_BINDINGS,
 	END_KEYS_BINDING
     };
     /* *INDENT-ON* */
@@ -538,6 +576,7 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
     int state = dialog_vars.defaultno ? dlg_defaultno_button() : sTEXT;
     int button = state;
     int first = (state == sTEXT);
+    int first_trace = TRUE;
     char *input;
     char *completed;
     char current[MAX_LEN + 1];
@@ -545,8 +584,8 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
     WINDOW *w_text = 0;
     WINDOW *w_work = 0;
     const char **buttons = dlg_ok_labels();
-    char *d_label = _("Directories");
-    char *f_label = _("Files");
+    const char *d_label = _("Directories");
+    const char *f_label = _("Files");
     char *partial;
     int min_wide = MIN_WIDE;
     int min_items = height ? 0 : 4;
@@ -556,9 +595,11 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
 
     /* Set up the initial value */
     input = dlg_set_result(path);
-    offset = strlen(input);
+    offset = (int) strlen(input);
     *current = 0;
 
+    dlg_button_layout(buttons, &min_wide);
+
 #ifdef KEY_RESIZE
   retry:
 #endif
@@ -577,8 +618,8 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
 
     dlg_mouse_setbase(0, 0);
 
-    dlg_draw_box(dialog, 0, 0, height, width, dialog_attr, border_attr);
-    dlg_draw_bottom_box(dialog);
+    dlg_draw_box2(dialog, 0, 0, height, width, dialog_attr, border_attr, border2_attr);
+    dlg_draw_bottom_box2(dialog, border_attr, border2_attr, dialog_attr);
     dlg_draw_title(dialog, title);
 
     wattrset(dialog, dialog_attr);
@@ -596,13 +637,15 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
     (void) keypad(w_text, TRUE);
     dlg_draw_box(dialog, tbox_y - MARGIN, tbox_x - MARGIN,
 		 (2 * MARGIN + 1), tbox_width + (MARGIN + EXT_WIDE),
-		 menubox_border_attr, menubox_attr);
+		 menubox_border_attr, menubox_border2_attr);
     dlg_mouse_mkbigregion(getbegy(dialog) + tbox_y - MARGIN,
 			  getbegx(dialog) + tbox_x - MARGIN,
 			  1 + (2 * MARGIN),
 			  tbox_width + (MARGIN + EXT_WIDE),
 			  MOUSE_T, 1, 1, 3 /* doesn't matter */ );
 
+    dlg_register_window(w_text, "fselect2", binding2);
+
     /* Draw the directory listing box */
     if (dselect)
 	dbox_width = (width - (6 * MARGIN));
@@ -621,7 +664,7 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
     dlg_draw_box(dialog,
 		 dbox_y - MARGIN, dbox_x - MARGIN,
 		 dbox_height + (MARGIN + 1), dbox_width + (MARGIN + 1),
-		 menubox_border_attr, menubox_attr);
+		 menubox_border_attr, menubox_border2_attr);
     init_list(&d_list, dialog, w_work, MOUSE_D);
 
     if (!dselect) {
@@ -640,7 +683,7 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
 	dlg_draw_box(dialog,
 		     fbox_y - MARGIN, fbox_x - MARGIN,
 		     fbox_height + (MARGIN + 1), fbox_width + (MARGIN + 1),
-		     menubox_border_attr, menubox_attr);
+		     menubox_border_attr, menubox_border2_attr);
 	init_list(&f_list, dialog, w_work, MOUSE_F);
     } else {
 	memset(&f_list, 0, sizeof(f_list));
@@ -655,7 +698,7 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
 	if (resized) {
 	    resized = FALSE;
 	    dlg_show_string(w_text, input, offset, inputbox_attr,
-			    0, 0, tbox_width, 0, first);
+			    0, 0, tbox_width, (bool) 0, (bool) first);
 	}
 #endif
 
@@ -667,6 +710,12 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
 	    button = (state < 0) ? 0 : state;
 	    dlg_draw_buttons(dialog, height - 2, 0, buttons, button, FALSE, width);
 	}
+
+	if (first_trace) {
+	    first_trace = FALSE;
+	    dlg_trace_win(dialog);
+	}
+
 	if (state < 0) {
 	    switch (state) {
 	    case sTEXT:
@@ -686,7 +735,7 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
 	} else {
 	    fix_arrows(&d_list);
 	    fix_arrows(&f_list);
-	    key = dlg_mouse_wgetch(dialog, &fkey);
+	    key = dlg_mouse_wgetch((state == sTEXT) ? w_text : dialog, &fkey);
 	    if (dlg_result_key(key, fkey, &result))
 		break;
 	}
@@ -756,7 +805,7 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
 		    state = sTEXT;
 		    show_buttons = TRUE;
 		    strcpy(leaf_of(input), completed);
-		    offset = strlen(input);
+		    offset = (int) strlen(input);
 		    dlg_show_string(w_text, input, offset, inputbox_attr,
 				    0, 0, tbox_width, 0, first);
 		    if (partial != NULL)
@@ -768,7 +817,7 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
 		}
 		/* FALLTHRU */
 	    case DLGK_ENTER:
-		result = (state > 0) ? dlg_ok_buttoncode(state) : DLG_EXIT_OK;
+		result = (state > 0) ? dlg_enter_buttoncode(state) : DLG_EXIT_OK;
 		continue;
 #ifdef KEY_RESIZE
 	    case KEY_RESIZE:
@@ -828,6 +877,7 @@ dlg_fselect(const char *title, const char *path, int height, int width, bool dse
 	}
     }
 
+    dlg_unregister_window(w_text);
     dlg_del_window(dialog);
     dlg_mouse_free_regions();
     free_list(&d_list, FALSE);
