@@ -1,9 +1,9 @@
 /*
- *  $Id: menubox.c,v 1.112 2008/03/16 12:54:58 tom Exp $
+ *  $Id: menubox.c,v 1.127 2011/10/17 00:49:25 tom Exp $
  *
  *  menubox.c -- implements the menu box
  *
- *  Copyright 2000-2007,2008	Thomas E. Dickey
+ *  Copyright 2000-2010,2011	Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public Licens, version 2.1e
@@ -57,11 +57,17 @@ print_arrows(WINDOW *win,
 	     int item_no,
 	     int menu_height)
 {
-    dlg_draw_arrows(win, scrollamt,
-		    scrollamt + max_choice < item_no,
-		    box_x + tag_x + 1,
-		    box_y,
-		    box_y + menu_height + 1);
+    dlg_draw_scrollbar(win,
+		       scrollamt,
+		       scrollamt,
+		       scrollamt + max_choice,
+		       item_no,
+		       box_x,
+		       box_x + menu_width,
+		       box_y,
+		       box_y + menu_height + 1,
+		       menubox_border2_attr,
+		       menubox_border_attr);
 }
 
 /*
@@ -77,15 +83,13 @@ print_tag(WINDOW *win,
     int my_x = item_x;
     int my_y = ItemToRow(choice);
     int tag_width = (my_x - tag_x - GUTTER);
-    const int *cols;
     const int *indx;
     int limit;
-    unsigned prefix;
+    int prefix;
 
-    cols = dlg_index_columns(item->name);
     indx = dlg_index_wchars(item->name);
     limit = dlg_count_wchars(item->name);
-    prefix = indx[1] - indx[0];
+    prefix = (indx[1] - indx[0]);
 
     /* highlight first char of the tag to be special */
     (void) wmove(win, my_y, tag_x);
@@ -94,7 +98,7 @@ print_tag(WINDOW *win,
 	(void) waddnstr(win, item->name, prefix);
     /* print rest of the string */
     wattrset(win, selected ? tag_selected_attr : tag_attr);
-    if (strlen(item->name) > prefix) {
+    if ((int) strlen(item->name) > prefix) {
 	limit = dlg_limit_columns(item->name, tag_width, 1);
 	if (limit > 0)
 	    (void) waddnstr(win, item->name + indx[1], indx[limit] - indx[1]);
@@ -111,7 +115,7 @@ print_item(WINDOW *win,
 	   Mode selected,
 	   bool is_inputmenu)
 {
-    chtype save = getattrs(win);
+    chtype save = dlg_get_attrs(win);
     int n;
     int my_width = menu_width;
     int my_x = item_x;
@@ -182,7 +186,7 @@ input_menu_edit(WINDOW *win,
 		int choice,
 		char **resultp)
 {
-    chtype save = getattrs(win);
+    chtype save = dlg_get_attrs(win);
     char *result;
     int offset = 0;
     int key = 0, fkey = 0;
@@ -191,9 +195,9 @@ input_menu_edit(WINDOW *win,
     bool is_inputmenu = TRUE;
     int y = ItemToRow(choice);
     int code = TRUE;
-    int max_len = dlg_max_input(MAX(strlen(items->text) + 1, MAX_LEN));
+    int max_len = dlg_max_input(MAX((int) strlen(items->text) + 1, MAX_LEN));
 
-    result = dlg_malloc(char, max_len);
+    result = dlg_malloc(char, (size_t) max_len);
     assert_ptr(result, "input_menu_edit");
 
     /* original item is used to initialize the input string. */
@@ -231,34 +235,34 @@ handle_button(int code, DIALOG_LISTITEM * items, int choice)
     switch (code) {
     case DLG_EXIT_OK:		/* FALLTHRU */
     case DLG_EXIT_EXTRA:
-	dlg_add_result(items[choice].name);
+	dlg_add_string(items[choice].name);
 	break;
     case DLG_EXIT_HELP:
 	dlg_add_result("HELP ");
 	if (USE_ITEM_HELP(items[choice].help)) {
-	    dlg_add_result(items[choice].help);
+	    dlg_add_string(items[choice].help);
 	    code = DLG_EXIT_ITEM_HELP;
 	} else {
-	    dlg_add_result(items[choice].name);
+	    dlg_add_string(items[choice].name);
 	}
 	break;
     }
     return code;
 }
 
-static int
+int
 dlg_renamed_menutext(DIALOG_LISTITEM * items, int current, char *newtext)
 {
     if (dialog_vars.input_result)
 	dialog_vars.input_result[0] = '\0';
     dlg_add_result("RENAMED ");
-    dlg_add_result(items[current].name);
+    dlg_add_string(items[current].name);
     dlg_add_result(" ");
-    dlg_add_result(newtext);
+    dlg_add_string(newtext);
     return DLG_EXIT_EXTRA;
 }
 
-static int
+int
 dlg_dummy_menutext(DIALOG_LISTITEM * items, int current, char *newtext)
 {
     (void) items;
@@ -285,6 +289,7 @@ dlg_menu(const char *title,
 {
     /* *INDENT-OFF* */
     static DLG_KEYS_BINDING binding[] = {
+	HELPKEY_BINDINGS,
 	ENTERKEY_BINDINGS,
 	DLG_KEYS_DATA( DLGK_FIELD_NEXT,	' ' ),
 	DLG_KEYS_DATA( DLGK_FIELD_NEXT,	KEY_RIGHT ),
@@ -306,6 +311,7 @@ dlg_menu(const char *title,
     };
     static DLG_KEYS_BINDING binding2[] = {
 	INPUTSTR_BINDINGS,
+	HELPKEY_BINDINGS,
 	ENTERKEY_BINDINGS,
 	END_KEYS_BINDING
     };
@@ -327,7 +333,8 @@ dlg_menu(const char *title,
     WINDOW *dialog, *menu;
     char *prompt = dlg_strclone(cprompt);
     const char **buttons = dlg_ok_labels();
-    bool is_inputmenu = (rename_menutext == dlg_renamed_menutext);
+    bool is_inputmenu = ((rename_menutext != 0)
+			 && (rename_menutext != dlg_dummy_menutext));
 
     dlg_does_output();
     dlg_tab_correct_str(prompt);
@@ -345,6 +352,7 @@ dlg_menu(const char *title,
     } else {
 	dlg_auto_size(title, prompt, &height, &width, MIN_HIGH + use_height, 26);
     }
+    dlg_button_layout(buttons, &width);
     dlg_print_size(height, width);
     dlg_ctl_size(height, width);
 
@@ -357,8 +365,8 @@ dlg_menu(const char *title,
 
     dlg_mouse_setbase(x, y);
 
-    dlg_draw_box(dialog, 0, 0, height, width, dialog_attr, border_attr);
-    dlg_draw_bottom_box(dialog);
+    dlg_draw_box2(dialog, 0, 0, height, width, dialog_attr, border_attr, border2_attr);
+    dlg_draw_bottom_box2(dialog, border_attr, border2_attr, dialog_attr);
     dlg_draw_title(dialog, title);
 
     wattrset(dialog, dialog_attr);
@@ -393,7 +401,7 @@ dlg_menu(const char *title,
 
     /* draw a box around the menu items */
     dlg_draw_box(dialog, box_y, box_x, use_height + 2, menu_width + 2,
-		 menubox_border_attr, menubox_attr);
+		 menubox_border_attr, menubox_border2_attr);
 
     name_width = 0;
     text_width = 0;
@@ -452,6 +460,7 @@ dlg_menu(const char *title,
 
     dlg_draw_buttons(dialog, height - 2, 0, buttons, button, FALSE, width);
 
+    dlg_trace_win(dialog);
     while (result == DLG_EXIT_UNKNOWN) {
 	if (button < 0)		/* --visit-items */
 	    wmove(dialog, box_y + ItemToRow(choice) + 1, box_x + tag_x + 1);
@@ -655,6 +664,9 @@ dlg_menu(const char *title,
 			       Selected,
 			       is_inputmenu);
 		    (void) wnoutrefresh(menu);
+		    print_arrows(dialog,
+				 box_x, box_y,
+				 scrollamt, max_choice, item_no, use_height);
 		    (void) wmove(dialog, cur_y, cur_x);
 		    wrefresh(dialog);
 		}
@@ -675,22 +687,15 @@ dlg_menu(const char *title,
 				 FALSE, width);
 		break;
 	    case DLGK_ENTER:
-		result = dlg_ok_buttoncode(button);
+		result = dlg_enter_buttoncode(button);
 
 		/*
 		 * If dlg_menu() is called from dialog_menu(), we want to
-		 * capture the results into dialog_vars.input_result, but not
-		 * if dlg_menu() is called directly from an application.  We
-		 * can check this by testing if rename_menutext is the function
-		 * pointer owned by dialog_menu().  It would be nicer to have
-		 * this logic inside dialog_menu(), but that cannot be done
-		 * since we would lose compatibility for the results reported
-		 * after input_menu_edit().
+		 * capture the results into dialog_vars.input_result.
 		 */
 		if (result == DLG_EXIT_ERROR) {
 		    result = DLG_EXIT_UNKNOWN;
-		} else if (is_inputmenu
-			   || rename_menutext == dlg_dummy_menutext) {
+		} else if (is_inputmenu) {
 		    result = handle_button(result,
 					   items,
 					   scrollamt + choice);
@@ -780,15 +785,17 @@ dialog_menu(const char *title,
     int i;
     DIALOG_LISTITEM *listitems;
 
-    listitems = dlg_calloc(DIALOG_LISTITEM, item_no + 1);
+    listitems = dlg_calloc(DIALOG_LISTITEM, (size_t) item_no + 1);
     assert_ptr(listitems, "dialog_menu");
 
     for (i = 0; i < item_no; ++i) {
 	listitems[i].name = ItemName(i);
 	listitems[i].text = ItemText(i);
-	listitems[i].help = (dialog_vars.item_help) ? ItemHelp(i) :
-	    dlg_strempty();
+	listitems[i].help = ((dialog_vars.item_help)
+			     ? ItemHelp(i)
+			     : dlg_strempty());
     }
+    dlg_align_columns(&listitems[0].text, sizeof(DIALOG_LISTITEM), item_no);
 
     result = dlg_menu(title,
 		      cprompt,
@@ -798,8 +805,11 @@ dialog_menu(const char *title,
 		      item_no,
 		      listitems,
 		      &choice,
-		      dialog_vars.input_menu ? dlg_renamed_menutext : dlg_dummy_menutext);
+		      (dialog_vars.input_menu
+		       ? dlg_renamed_menutext
+		       : dlg_dummy_menutext));
 
+    dlg_free_columns(&listitems[0].text, sizeof(DIALOG_LISTITEM), item_no);
     free(listitems);
     return result;
 }
