@@ -1,9 +1,9 @@
 /*
- *  $Id: util.c,v 1.187 2008/03/16 18:46:19 tom Exp $
+ *  $Id: util.c,v 1.239 2011/10/18 22:25:13 tom Exp $
  *
  *  util.c -- miscellaneous utilities for dialog
  *
- *  Copyright 2000-2007,2008	Thomas E. Dickey
+ *  Copyright 2000-2010,2011	Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License, version 2.1
@@ -27,6 +27,14 @@
 #include <dialog.h>
 #include <dlg_keys.h>
 
+#ifdef HAVE_SETLOCALE
+#include <locale.h>
+#endif
+
+#ifdef NEED_WCHAR_H
+#include <wchar.h>
+#endif
+
 #ifdef NCURSES_VERSION
 #if defined(HAVE_NCURSESW_TERM_H)
 #include <ncursesw/term.h>
@@ -37,10 +45,30 @@
 #endif
 #endif
 
+#if defined(HAVE_WCHGAT)
+#  if defined(NCURSES_VERSION_PATCH)
+#    if NCURSES_VERSION_PATCH >= 20060715
+#      define USE_WCHGAT 1
+#    else
+#      define USE_WCHGAT 0
+#    endif
+#  else
+#    define USE_WCHGAT 1
+#  endif
+#else
+#  define USE_WCHGAT 0
+#endif
+
 /* globals */
 DIALOG_STATE dialog_state;
 DIALOG_VARS dialog_vars;
 
+#if !(defined(HAVE_WGETPARENT) && defined(HAVE_WINDOW__PARENT))
+#define NEED_WGETPARENT 1
+#else
+#undef NEED_WGETPARENT
+#endif
+
 #define concat(a,b) a##b
 
 #ifdef HAVE_RC_FILE
@@ -61,8 +89,11 @@ DIALOG_VARS dialog_vars;
 
 #define DATA(atr,upr,lwr,cmt) { atr COLOR_DATA(upr) RC_DATA(lwr,cmt) }
 
+#define UseShadow(dw) ((dw) != 0 && (dw)->normal != 0 && (dw)->shadow != 0)
+
 /*
  * Table of color and attribute values, default is for mono display.
+ * The order matches the DIALOG_ATR() values.
  */
 /* *INDENT-OFF* */
 DIALOG_COLORS dlg_color_table[] =
@@ -99,10 +130,51 @@ DIALOG_COLORS dlg_color_table[] =
     DATA(A_NORMAL,	ITEMHELP,		itemhelp, "Item help-text"),
     DATA(A_BOLD,	FORM_ACTIVE_TEXT,	form_active_text, "Active form text"),
     DATA(A_REVERSE,	FORM_TEXT,		form_text, "Form text"),
-    DATA(A_NORMAL,	FORM_ITEM_READONLY,	form_item_readonly, "Readonly form item") 
+    DATA(A_NORMAL,	FORM_ITEM_READONLY,	form_item_readonly, "Readonly form item"),
+    DATA(A_REVERSE,	GAUGE,			gauge, "Dialog box gauge"),
+    DATA(A_REVERSE,	BORDER2,		border2, "Dialog box border2"),
+    DATA(A_REVERSE,	INPUTBOX_BORDER2,	inputbox_border2, "Input box border2"),
+    DATA(A_REVERSE,	SEARCHBOX_BORDER2,	searchbox_border2, "Search box border2"),
+    DATA(A_REVERSE,	MENUBOX_BORDER2,	menubox_border2, "Menu box border2")
 };
 /* *INDENT-ON* */
 
+/*
+ * Maintain a list of subwindows so that we can delete them to cleanup.
+ * More important, this provides a fallback when wgetparent() is not available.
+ */
+static void
+add_subwindow(WINDOW *parent, WINDOW *child)
+{
+    DIALOG_WINDOWS *p = dlg_calloc(DIALOG_WINDOWS, 1);
+
+    if (p != 0) {
+	p->normal = parent;
+	p->shadow = child;
+	p->next = dialog_state.all_subwindows;
+	dialog_state.all_subwindows = p;
+    }
+}
+
+static void
+del_subwindows(WINDOW *parent)
+{
+    DIALOG_WINDOWS *p, *q;
+
+    for (p = dialog_state.all_subwindows, q = 0; p != 0; q = p, p = p->next) {
+	if (p->normal == parent) {
+	    delwin(p->shadow);
+	    if (q == 0) {
+		dialog_state.all_subwindows = p->next;
+	    } else {
+		q->next = p->next;
+	    }
+	    free(p);
+	    continue;
+	}
+    }
+}
+
 /*
  * Display background title if it exists ...
  */
@@ -209,6 +281,8 @@ init_dialog(FILE *input, FILE *output)
     int fd1, fd2;
     char *device = 0;
 
+    setlocale(LC_ALL, "");
+
     dialog_state.output = output;
     dialog_state.tab_len = TAB_LEN;
     dialog_state.aspect_ratio = DEFAULT_ASPECT_RATIO;
@@ -327,7 +401,11 @@ init_dialog(FILE *input, FILE *output)
     (void) keypad(stdscr, TRUE);
     (void) cbreak();
     (void) noecho();
-    mouse_open();
+
+    if (!dialog_state.no_mouse) {
+	mouse_open();
+    }
+
     dialog_state.screen_initialized = TRUE;
 
 #ifdef HAVE_COLOR
@@ -398,6 +476,24 @@ dlg_color_count(void)
     return sizeof(dlg_color_table) / sizeof(dlg_color_table[0]);
 }
 
+/*
+ * Wrapper for getattrs(), or the more cumbersome X/Open wattr_get().
+ */
+chtype
+dlg_get_attrs(WINDOW *win)
+{
+    chtype result;
+#ifdef HAVE_GETATTRS
+    result = (chtype) getattrs(win);
+#else
+    attr_t my_result;
+    short my_pair;
+    wattr_get(win, &my_result, &my_pair, NULL);
+    result = my_result;
+#endif
+    return result;
+}
+
 /*
  * Reuse color pairs (they are limited), returning a COLOR_PAIR() value if we
  * have (or can) define a pair with the given color as foreground on the
@@ -412,18 +508,18 @@ dlg_color_pair(int foreground, int background)
     bool found = FALSE;
 
     for (pair = 1; pair < defined_colors; ++pair) {
-	if (pair_content(pair, &fg, &bg) != ERR
+	if (pair_content((short) pair, &fg, &bg) != ERR
 	    && fg == foreground
 	    && bg == background) {
-	    result = COLOR_PAIR(pair);
+	    result = (chtype) COLOR_PAIR(pair);
 	    found = TRUE;
 	    break;
 	}
     }
     if (!found && (defined_colors + 1) < COLOR_PAIRS) {
 	pair = defined_colors++;
-	(void) init_pair(pair, foreground, background);
-	result = COLOR_PAIR(pair);
+	(void) init_pair((short) pair, (short) foreground, (short) background);
+	result = (chtype) COLOR_PAIR(pair);
     }
     return result;
 }
@@ -436,12 +532,12 @@ dlg_color_pair(int foreground, int background)
 static chtype
 define_color(WINDOW *win, int foreground)
 {
-    chtype attrs = getattrs(win);
+    chtype attrs = dlg_get_attrs(win);
     int pair;
     short fg, bg, background;
 
     if ((pair = PAIR_NUMBER(attrs)) != 0
-	&& pair_content(pair, &fg, &bg) != ERR) {
+	&& pair_content((short) pair, &fg, &bg) != ERR) {
 	background = bg;
     } else {
 	background = COLOR_BLACK;
@@ -464,29 +560,68 @@ end_dialog(void)
     }
 }
 
+#define ESCAPE_LEN 3
 #define isOurEscape(p) (((p)[0] == '\\') && ((p)[1] == 'Z') && ((p)[2] != 0))
 
+int
+dlg_count_real_columns(const char *text)
+{
+    int result = dlg_count_columns(text);
+    if (result && dialog_vars.colors) {
+	int hidden = 0;
+	while (*text) {
+	    if (dialog_vars.colors && isOurEscape(text)) {
+		hidden += ESCAPE_LEN;
+		text += ESCAPE_LEN;
+	    } else {
+		++text;
+	    }
+	}
+	result -= hidden;
+    }
+    return result;
+}
+
 static int
 centered(int width, const char *string)
 {
-    int len = dlg_count_columns(string);
+    int need = dlg_count_real_columns(string);
     int left;
-    int hide = 0;
-    int n;
 
-    if (dialog_vars.colors) {
-	for (n = 0; n < len; ++n) {
-	    if (isOurEscape(string + n)) {
-		hide += 3;
-	    }
-	}
-    }
-    left = (width - (len - hide)) / 2 - 1;
+    left = (width - need) / 2 - 1;
     if (left < 0)
 	left = 0;
     return left;
 }
 
+#ifdef USE_WIDE_CURSES
+static bool
+is_combining(const char *txt, int *combined)
+{
+    bool result = FALSE;
+
+    if (*combined == 0) {
+	if (UCH(*txt) >= 128) {
+	    wchar_t wch;
+	    mbstate_t state;
+	    size_t given = strlen(txt);
+	    size_t len;
+
+	    memset(&state, 0, sizeof(state));
+	    len = mbrtowc(&wch, txt, given, &state);
+	    if ((int) len > 0 && wcwidth(wch) == 0) {
+		*combined = (int) len - 1;
+		result = TRUE;
+	    }
+	}
+    } else {
+	result = TRUE;
+	*combined -= 1;
+    }
+    return result;
+}
+#endif
+
 /*
  * Print up to 'cols' columns from 'text', optionally rendering our escape
  * sequence for attributes and color.
@@ -501,6 +636,9 @@ dlg_print_text(WINDOW *win, const char *txt, int cols, chtype *attr)
     bool thisTab;
     bool ended = FALSE;
     chtype useattr;
+#ifdef USE_WIDE_CURSES
+    int combined = 0;
+#endif
 
     getyx(win, y_origin, x_origin);
     while (cols > 0 && (*txt != '\0')) {
@@ -558,7 +696,7 @@ dlg_print_text(WINDOW *win, const char *txt, int cols, chtype *attr)
 	 * attribute.
 	 */
 	if ((useattr & A_COLOR) != 0 && (useattr & A_BOLD) == 0) {
-	    short pair = PAIR_NUMBER(useattr);
+	    short pair = (short) PAIR_NUMBER(useattr);
 	    short fg, bg;
 	    if (pair_content(pair, &fg, &bg) != ERR
 		&& fg == bg) {
@@ -576,13 +714,20 @@ dlg_print_text(WINDOW *win, const char *txt, int cols, chtype *attr)
 	 * more blanks.
 	 */
 	thisTab = (CharOf(*txt) == TAB);
-	if (thisTab)
+	if (thisTab) {
 	    getyx(win, y_before, x_before);
+	    (void) y_before;
+	}
 	(void) waddch(win, CharOf(*txt++) | useattr);
 	getyx(win, y_after, x_after);
 	if (thisTab && (y_after == y_origin))
 	    tabbed += (x_after - x_before);
-	if (y_after != y_origin || x_after >= cols + tabbed + x_origin) {
+	if ((y_after != y_origin) ||
+	    (x_after >= (cols + tabbed + x_origin)
+#ifdef USE_WIDE_CURSES
+	     && !is_combining(txt, &combined)
+#endif
+	    )) {
 	    ended = TRUE;
 	}
     }
@@ -602,6 +747,7 @@ dlg_print_line(WINDOW *win,
 {
     const char *wrap_ptr = prompt;
     const char *test_ptr = prompt;
+    const char *hide_ptr = 0;
     const int *cols = dlg_index_columns(prompt);
     const int *indx = dlg_index_wchars(prompt);
     int wrap_inx = 0;
@@ -627,9 +773,10 @@ dlg_print_line(WINDOW *win,
 	} else if (*test_ptr == ' ' && n != 0 && prompt[indx[n - 1]] != ' ') {
 	    wrap_inx = n;
 	    *x = cur_x;
-	} else if (isOurEscape(test_ptr)) {
-	    hidden += 3;
-	    n += 2;
+	} else if (dialog_vars.colors && isOurEscape(test_ptr)) {
+	    hide_ptr = test_ptr;
+	    hidden += ESCAPE_LEN;
+	    n += (ESCAPE_LEN - 1);
 	}
 	cur_x = lm + tabbed + cols[n + 1];
 	if (cur_x > (rm + hidden))
@@ -656,6 +803,31 @@ dlg_print_line(WINDOW *win,
 	wrap_inx = test_inx;
     }
     wrap_ptr = prompt + indx[wrap_inx];
+#ifdef USE_WIDE_CURSES
+    if (UCH(*wrap_ptr) >= 128) {
+	int combined = 0;
+	while (is_combining(wrap_ptr, &combined)) {
+	    ++wrap_ptr;
+	}
+    }
+#endif
+
+    /*
+     * If we found hidden text past the last point that we will display,
+     * discount that from the displayed length.
+     */
+    if ((hide_ptr != 0) && (hide_ptr >= wrap_ptr)) {
+	hidden -= ESCAPE_LEN;
+	test_ptr = wrap_ptr;
+	while (test_ptr < wrap_ptr) {
+	    if (dialog_vars.colors && isOurEscape(test_ptr)) {
+		hidden -= ESCAPE_LEN;
+		test_ptr += ESCAPE_LEN;
+	    } else {
+		++test_ptr;
+	    }
+	}
+    }
 
     /*
      * Print the line if we have a window pointer.  Otherwise this routine
@@ -669,6 +841,8 @@ dlg_print_line(WINDOW *win,
     if (*x == 1)
 	*x = rm;
 
+    *x -= hidden;
+
     /* Find the start of the next line and return a pointer to it */
     test_ptr = wrap_ptr;
     while (*test_ptr == ' ')
@@ -755,6 +929,158 @@ dlg_print_autowrap(WINDOW *win, const char *prompt, int height, int width)
 		 (int *) 0, (int *) 0);
 }
 
+/*
+ * Display the message in a scrollable window.  Actually the way it works is
+ * that we create a "tall" window of the proper width, let the text wrap within
+ * that, and copy a slice of the result to the dialog.
+ *
+ * It works for ncurses.  Other curses implementations show only blanks (Tru64)
+ * or garbage (NetBSD).
+ */
+int
+dlg_print_scrolled(WINDOW *win,
+		   const char *prompt,
+		   int offset,
+		   int height,
+		   int width,
+		   int pauseopt)
+{
+    int oldy, oldx;
+    int last = 0;
+
+    (void) pauseopt;		/* used only for ncurses */
+
+    getyx(win, oldy, oldx);
+#ifdef NCURSES_VERSION
+    if (pauseopt) {
+	int wide = width - (2 * MARGIN);
+	int high = LINES;
+	int y, x;
+	int len;
+	int percent;
+	WINDOW *dummy;
+	char buffer[5];
+
+#if defined(NCURSES_VERSION_PATCH) && NCURSES_VERSION_PATCH >= 20040417
+	/*
+	 * If we're not limited by the screensize, allow text to possibly be
+	 * one character per line.
+	 */
+	if ((len = dlg_count_columns(prompt)) > high)
+	    high = len;
+#endif
+	dummy = newwin(high, width, 0, 0);
+	if (dummy == 0) {
+	    wattrset(win, dialog_attr);
+	    dlg_print_autowrap(win, prompt, height + 1 + (3 * MARGIN), width);
+	    last = 0;
+	} else {
+	    wbkgdset(dummy, dialog_attr | ' ');
+	    wattrset(dummy, dialog_attr);
+	    werase(dummy);
+	    dlg_print_autowrap(dummy, prompt, high, width);
+	    getyx(dummy, y, x);
+	    (void) x;
+
+	    copywin(dummy,	/* srcwin */
+		    win,	/* dstwin */
+		    offset + MARGIN,	/* sminrow */
+		    MARGIN,	/* smincol */
+		    MARGIN,	/* dminrow */
+		    MARGIN,	/* dmincol */
+		    height,	/* dmaxrow */
+		    wide,	/* dmaxcol */
+		    FALSE);
+
+	    delwin(dummy);
+
+	    /* if the text is incomplete, or we have scrolled, show the percentage */
+	    if (y > 0 && wide > 4) {
+		percent = (int) ((height + offset) * 100.0 / y);
+		if (percent < 0)
+		    percent = 0;
+		if (percent > 100)
+		    percent = 100;
+		if (offset != 0 || percent != 100) {
+		    (void) wattrset(win, position_indicator_attr);
+		    (void) wmove(win, MARGIN + height, wide - 4);
+		    (void) sprintf(buffer, "%d%%", percent);
+		    (void) waddstr(win, buffer);
+		    if ((len = (int) strlen(buffer)) < 4) {
+			wattrset(win, border_attr);
+			whline(win, dlg_boxchar(ACS_HLINE), 4 - len);
+		    }
+		}
+	    }
+	    last = (y - height);
+	}
+    } else
+#endif
+    {
+	(void) offset;
+	wattrset(win, dialog_attr);
+	dlg_print_autowrap(win, prompt, height + 1 + (3 * MARGIN), width);
+	last = 0;
+    }
+    wmove(win, oldy, oldx);
+    return last;
+}
+
+int
+dlg_check_scrolled(int key, int last, int page, bool * show, int *offset)
+{
+    int code = 0;
+
+    *show = FALSE;
+
+    switch (key) {
+    case DLGK_PAGE_FIRST:
+	if (*offset > 0) {
+	    *offset = 0;
+	    *show = TRUE;
+	}
+	break;
+    case DLGK_PAGE_LAST:
+	if (*offset < last) {
+	    *offset = last;
+	    *show = TRUE;
+	}
+	break;
+    case DLGK_GRID_UP:
+	if (*offset > 0) {
+	    --(*offset);
+	    *show = TRUE;
+	}
+	break;
+    case DLGK_GRID_DOWN:
+	if (*offset < last) {
+	    ++(*offset);
+	    *show = TRUE;
+	}
+	break;
+    case DLGK_PAGE_PREV:
+	if (*offset > 0) {
+	    *offset -= page;
+	    if (*offset < 0)
+		*offset = 0;
+	    *show = TRUE;
+	}
+	break;
+    case DLGK_PAGE_NEXT:
+	if (*offset < last) {
+	    *offset += page;
+	    if (*offset > last)
+		*offset = last;
+	    *show = TRUE;
+	}
+	break;
+    default:
+	code = -1;
+	break;
+    }
+    return code;
+}
+
 /*
  * Calculate the window size for preformatted text.  This will calculate box
  * dimensions that are at or close to the specified aspect ratio for the prompt
@@ -782,9 +1108,9 @@ auto_size_preformatted(const char *prompt, int *height, int *width)
      */
     if (car > ar) {
 	diff = car / (float) ar;
-	max_x = (int) (wide / diff + 4);
+	max_x = (int) ((float) wide / diff + 4);
 	justify_text((WINDOW *) 0, prompt, max_y, max_x, &high, &wide);
-	car = (float) wide / high;
+	car = (float) wide / (float) high;
     }
 
     /*
@@ -862,7 +1188,7 @@ real_auto_size(const char *title,
     } else if (prompt != 0) {
 	wide = MAX(title_length, mincols);
 	if (strchr(prompt, '\n') == 0) {
-	    double val = dialog_state.aspect_ratio * dlg_count_columns(prompt);
+	    double val = dialog_state.aspect_ratio * dlg_count_real_columns(prompt);
 	    double xxx = sqrt(val);
 	    int tmp = (int) xxx;
 	    wide = MAX(wide, tmp);
@@ -962,12 +1288,12 @@ dlg_auto_sizefile(const char *title,
 		offset++;
 
 	if (offset > len)
-	    len = offset;
+	    len = (int) offset;
 
 	count++;
     }
 
-    /* now 'count' has the number of lines of fd and 'len' the max lenght */
+    /* now 'count' has the number of lines of fd and 'len' the max length */
 
     *height = MIN(SLINES, count + numlines + boxlines);
     *width = MIN(SCOLS, MAX((len + nc), mincols));
@@ -978,6 +1304,27 @@ dlg_auto_sizefile(const char *title,
     (void) fclose(fd);
 }
 
+static chtype
+dlg_get_cell_attrs(WINDOW *win)
+{
+    chtype result;
+#ifdef USE_WIDE_CURSES
+    cchar_t wch;
+    wchar_t cc;
+    attr_t attrs;
+    short pair;
+    if (win_wch(win, &wch) == OK
+	&& getcchar(&wch, &cc, &attrs, &pair, NULL) == OK) {
+	result = attrs;
+    } else {
+	result = 0;
+    }
+#else
+    result = winch(win) & (A_ATTRIBUTES & ~A_COLOR);
+#endif
+    return result;
+}
+
 /*
  * Draw a rectangular box with line drawing characters.
  *
@@ -995,11 +1342,11 @@ dlg_auto_sizefile(const char *title,
  * reverse this choice.
  */
 void
-dlg_draw_box(WINDOW *win, int y, int x, int height, int width,
-	     chtype boxchar, chtype borderchar)
+dlg_draw_box2(WINDOW *win, int y, int x, int height, int width,
+	      chtype boxchar, chtype borderchar, chtype borderchar2)
 {
     int i, j;
-    chtype save = getattrs(win);
+    chtype save = dlg_get_attrs(win);
 
     wattrset(win, 0);
     for (i = 0; i < height; i++) {
@@ -1010,70 +1357,244 @@ dlg_draw_box(WINDOW *win, int y, int x, int height, int width,
 	    else if (i == height - 1 && !j)
 		(void) waddch(win, borderchar | dlg_boxchar(ACS_LLCORNER));
 	    else if (!i && j == width - 1)
-		(void) waddch(win, boxchar | dlg_boxchar(ACS_URCORNER));
+		(void) waddch(win, borderchar2 | dlg_boxchar(ACS_URCORNER));
 	    else if (i == height - 1 && j == width - 1)
-		(void) waddch(win, boxchar | dlg_boxchar(ACS_LRCORNER));
+		(void) waddch(win, borderchar2 | dlg_boxchar(ACS_LRCORNER));
 	    else if (!i)
 		(void) waddch(win, borderchar | dlg_boxchar(ACS_HLINE));
 	    else if (i == height - 1)
-		(void) waddch(win, boxchar | dlg_boxchar(ACS_HLINE));
+		(void) waddch(win, borderchar2 | dlg_boxchar(ACS_HLINE));
 	    else if (!j)
 		(void) waddch(win, borderchar | dlg_boxchar(ACS_VLINE));
 	    else if (j == width - 1)
-		(void) waddch(win, boxchar | dlg_boxchar(ACS_VLINE));
+		(void) waddch(win, borderchar2 | dlg_boxchar(ACS_VLINE));
 	    else
 		(void) waddch(win, boxchar | ' ');
     }
     wattrset(win, save);
 }
 
+void
+dlg_draw_box(WINDOW *win, int y, int x, int height, int width,
+	     chtype boxchar, chtype borderchar)
+{
+    dlg_draw_box2(win, y, x, height, width, boxchar, borderchar, boxchar);
+}
+
+static DIALOG_WINDOWS *
+find_window(WINDOW *win)
+{
+    DIALOG_WINDOWS *result = 0;
+    DIALOG_WINDOWS *p;
+
+    for (p = dialog_state.all_windows; p != 0; p = p->next) {
+	if (p->normal == win) {
+	    result = p;
+	    break;
+	}
+    }
+    return result;
+}
+
 #ifdef HAVE_COLOR
+/*
+ * If we have wchgat(), use that for updating shadow attributes, to work with
+ * wide-character data.
+ */
+
+/*
+ * Check if the given point is "in" the given window.  If so, return the window
+ * pointer, otherwise null.
+ */
+static WINDOW *
+in_window(WINDOW *win, int y, int x)
+{
+    WINDOW *result = 0;
+    int y_base = getbegy(win);
+    int x_base = getbegx(win);
+    int y_last = getmaxy(win) + y_base;
+    int x_last = getmaxx(win) + x_base;
+
+    if (y >= y_base && y <= y_last && x >= x_base && x <= x_last)
+	result = win;
+    return result;
+}
+
+static WINDOW *
+window_at_cell(DIALOG_WINDOWS * dw, int y, int x)
+{
+    WINDOW *result = 0;
+    DIALOG_WINDOWS *p;
+    int y_want = y + getbegy(dw->shadow);
+    int x_want = x + getbegx(dw->shadow);
+
+    for (p = dialog_state.all_windows; p != 0; p = p->next) {
+	if (dw->normal != p->normal
+	    && dw->shadow != p->normal
+	    && (result = in_window(p->normal, y_want, x_want)) != 0) {
+	    break;
+	}
+    }
+    if (result == 0) {
+	result = stdscr;
+    }
+    return result;
+}
+
+static bool
+in_shadow(WINDOW *normal, WINDOW *shadow, int y, int x)
+{
+    bool result = FALSE;
+    int ybase = getbegy(normal);
+    int ylast = getmaxy(normal) + ybase;
+    int xbase = getbegx(normal);
+    int xlast = getmaxx(normal) + xbase;
+
+    y += getbegy(shadow);
+    x += getbegx(shadow);
+
+    if (y >= ybase + SHADOW_ROWS
+	&& y < ylast + SHADOW_ROWS
+	&& x >= xlast
+	&& x < xlast + SHADOW_COLS) {
+	/* in the right-side */
+	result = TRUE;
+    } else if (y >= ylast
+	       && y < ylast + SHADOW_ROWS
+	       && x >= ybase + SHADOW_COLS
+	       && x < ylast + SHADOW_COLS) {
+	/* check the bottom */
+	result = TRUE;
+    }
+
+    return result;
+}
+
+/*
+ * When erasing a shadow, check each cell to make sure that it is not part of
+ * another box's shadow.  This is a little complicated since most shadows are
+ * merged onto stdscr.
+ */
+static bool
+last_shadow(DIALOG_WINDOWS * dw, int y, int x)
+{
+    DIALOG_WINDOWS *p;
+    bool result = TRUE;
+
+    for (p = dialog_state.all_windows; p != 0; p = p->next) {
+	if (p->normal != dw->normal
+	    && in_shadow(p->normal, dw->shadow, y, x)) {
+	    result = FALSE;
+	    break;
+	}
+    }
+    return result;
+}
+
+static void
+repaint_cell(DIALOG_WINDOWS * dw, bool draw, int y, int x)
+{
+    WINDOW *win = dw->shadow;
+    WINDOW *cellwin;
+    int y2, x2;
+
+    if ((cellwin = window_at_cell(dw, y, x)) != 0
+	&& (draw || last_shadow(dw, y, x))
+	&& (y2 = (y + getbegy(win) - getbegy(cellwin))) >= 0
+	&& (x2 = (x + getbegx(win) - getbegx(cellwin))) >= 0
+	&& wmove(cellwin, y2, x2) != ERR) {
+	chtype the_cell = dlg_get_attrs(cellwin);
+	chtype the_attr = (draw ? shadow_attr : the_cell);
+
+	if (dlg_get_cell_attrs(cellwin) & A_ALTCHARSET) {
+	    the_attr |= A_ALTCHARSET;
+	}
+#if USE_WCHGAT
+	wchgat(cellwin, 1,
+	       the_attr & (chtype) (~A_COLOR),
+	       (short) PAIR_NUMBER(the_attr),
+	       NULL);
+#else
+	{
+	    chtype the_char = ((winch(cellwin) & A_CHARTEXT) | the_attr);
+	    (void) waddch(cellwin, the_char);
+	}
+#endif
+	wnoutrefresh(cellwin);
+    }
+}
+
+#define RepaintCell(dw, draw, y, x) repaint_cell(dw, draw, y, x)
+
+static void
+repaint_shadow(DIALOG_WINDOWS * dw, bool draw, int y, int x, int height, int width)
+{
+    int i, j;
+
+    if (UseShadow(dw)) {
+#if !USE_WCHGAT
+	chtype save = dlg_get_attrs(dw->shadow);
+	wattrset(dw->shadow, draw ? shadow_attr : screen_attr);
+#endif
+	for (i = 0; i < SHADOW_ROWS; ++i) {
+	    for (j = 0; j < width; ++j) {
+		RepaintCell(dw, draw, i + y + height, j + x + SHADOW_COLS);
+	    }
+	}
+	for (i = 0; i < height; i++) {
+	    for (j = 0; j < SHADOW_COLS; ++j) {
+		RepaintCell(dw, draw, i + y + SHADOW_ROWS, j + x + width);
+	    }
+	}
+	(void) wnoutrefresh(dw->shadow);
+#if !USE_WCHGAT
+	wattrset(dw->shadow, save);
+#endif
+    }
+}
+
 /*
  * Draw a shadow on the parent window corresponding to the right- and
  * bottom-edge of the child window, to give a 3-dimensional look.
  */
 static void
-draw_childs_shadow(WINDOW *parent, WINDOW *child)
+draw_childs_shadow(DIALOG_WINDOWS * dw)
 {
-    if (has_colors()) {		/* Whether terminal supports color? */
-	chtype save = getattrs(parent);
+    if (UseShadow(dw)) {
+	repaint_shadow(dw,
+		       TRUE,
+		       getbegy(dw->normal) - getbegy(dw->shadow),
+		       getbegx(dw->normal) - getbegx(dw->shadow),
+		       getmaxy(dw->normal),
+		       getmaxx(dw->normal));
+    }
+}
 
-	dlg_draw_shadow(parent,
-			getbegy(child) - getbegy(parent),
-			getbegx(child) - getbegx(parent),
-			getmaxy(child),
-			getmaxx(child));
-	wattrset(parent, save);
+/*
+ * Erase a shadow on the parent window corresponding to the right- and
+ * bottom-edge of the child window.
+ */
+static void
+erase_childs_shadow(DIALOG_WINDOWS * dw)
+{
+    if (UseShadow(dw)) {
+	repaint_shadow(dw,
+		       FALSE,
+		       getbegy(dw->normal) - getbegy(dw->shadow),
+		       getbegx(dw->normal) - getbegx(dw->shadow),
+		       getmaxy(dw->normal),
+		       getmaxx(dw->normal));
     }
 }
 
 /*
  * Draw shadows along the right and bottom edge to give a more 3D look
- * to the boxes
+ * to the boxes.
  */
 void
 dlg_draw_shadow(WINDOW *win, int y, int x, int height, int width)
 {
-    int i, j;
-
-    if (has_colors()) {		/* Whether terminal supports color? */
-	wattrset(win, shadow_attr);
-	for (i = 0; i < SHADOW_ROWS; ++i) {
-	    for (j = 0; j < width; ++j) {
-		if (wmove(win, i + y + height, j + x + SHADOW_COLS) != ERR) {
-		    (void) waddch(win, winch(win) & (chtype) (~A_COLOR));
-		}
-	    }
-	}
-	for (i = 0; i < height; i++) {
-	    for (j = 0; j < SHADOW_COLS; ++j) {
-		if (wmove(win, i + y + SHADOW_ROWS, j + x + width) != ERR) {
-		    (void) waddch(win, winch(win) & (chtype) (~A_COLOR));
-		}
-	    }
-	}
-	(void) wnoutrefresh(win);
-    }
+    repaint_shadow(find_window(win), TRUE, y, x, height, width);
 }
 #endif /* HAVE_COLOR */
 
@@ -1111,7 +1632,7 @@ dlg_exit(int code)
 	    if ((name = getenv(table[n].name)) != 0) {
 		value = strtol(name, &temp, 0);
 		if (temp != 0 && temp != name && *temp == '\0') {
-		    code = value;
+		    code = (int) value;
 		    overridden = TRUE;
 		}
 	    }
@@ -1128,6 +1649,10 @@ dlg_exit(int code)
 	code = DLG_EXIT_HELP;
 	goto retry;
     }
+#ifdef HAVE_DLG_TRACE
+    dlg_trace((const char *) 0);	/* close it */
+#endif
+
 #ifdef NO_LEAKS
     _dlg_inputstr_leaks();
 #if defined(NCURSES_VERSION) && defined(HAVE__NC_FREE_AND_EXIT)
@@ -1146,9 +1671,11 @@ dlg_exit(int code)
 	    fclose(dialog_state.input);
 	    dialog_state.input = 0;
 	}
-	if (dialog_state.pipe_input != stdin) {
-	    fclose(dialog_state.pipe_input);
-	    dialog_state.pipe_input = 0;
+	if (dialog_state.pipe_input) {
+	    if (dialog_state.pipe_input != stdin) {
+		fclose(dialog_state.pipe_input);
+		dialog_state.pipe_input = 0;
+	    }
 	}
 	_exit(code);
     }
@@ -1364,7 +1891,7 @@ dlg_draw_title(WINDOW *win, const char *title)
 {
     if (title != NULL) {
 	chtype attr = A_NORMAL;
-	chtype save = getattrs(win);
+	chtype save = dlg_get_attrs(win);
 	int x = centered(getmaxx(win), title);
 
 	wattrset(win, title_attr);
@@ -1375,24 +1902,31 @@ dlg_draw_title(WINDOW *win, const char *title)
 }
 
 void
-dlg_draw_bottom_box(WINDOW *win)
+dlg_draw_bottom_box2(WINDOW *win, chtype on_left, chtype on_right, chtype on_inside)
 {
     int width = getmaxx(win);
     int height = getmaxy(win);
     int i;
 
-    wattrset(win, border_attr);
+    wattrset(win, on_left);
     (void) wmove(win, height - 3, 0);
     (void) waddch(win, dlg_boxchar(ACS_LTEE));
     for (i = 0; i < width - 2; i++)
 	(void) waddch(win, dlg_boxchar(ACS_HLINE));
-    wattrset(win, dialog_attr);
+    wattrset(win, on_right);
     (void) waddch(win, dlg_boxchar(ACS_RTEE));
+    wattrset(win, on_inside);
     (void) wmove(win, height - 2, 1);
     for (i = 0; i < width - 2; i++)
 	(void) waddch(win, ' ');
 }
 
+void
+dlg_draw_bottom_box(WINDOW *win)
+{
+    dlg_draw_bottom_box2(win, border_attr, dialog_attr, dialog_attr);
+}
+
 /*
  * Remove a window, repainting everything else.  This would be simpler if we
  * used the panel library, but that is not _always_ available.
@@ -1436,6 +1970,9 @@ dlg_del_window(WINDOW *win)
     }
 
     if (q) {
+	if (dialog_state.all_windows != 0)
+	    erase_childs_shadow(q);
+	del_subwindows(q->normal);
 	delwin(q->normal);
 	dlg_unregister_window(q->normal);
 	free(q);
@@ -1449,26 +1986,13 @@ dlg_del_window(WINDOW *win)
 WINDOW *
 dlg_new_window(int height, int width, int y, int x)
 {
-    WINDOW *win;
-    DIALOG_WINDOWS *p = dlg_calloc(DIALOG_WINDOWS, 1);
-
-    if ((win = newwin(height, width, y, x)) == 0) {
-	dlg_exiterr("Can't make new window at (%d,%d), size (%d,%d).\n",
-		    y, x, height, width);
-    }
-    p->next = dialog_state.all_windows;
-    p->normal = win;
-    dialog_state.all_windows = p;
-#ifdef HAVE_COLOR
-    if (dialog_state.use_shadow) {
-	draw_childs_shadow(p->shadow = stdscr, win);
-    }
-#endif
-
-    (void) keypad(win, TRUE);
-    return win;
+    return dlg_new_modal_window(stdscr, height, width, y, x);
 }
 
+/*
+ * "Modal" windows differ from normal ones by having a shadow in a window
+ * separate from the standard screen.
+ */
 WINDOW *
 dlg_new_modal_window(WINDOW *parent, int height, int width, int y, int x)
 {
@@ -1485,7 +2009,8 @@ dlg_new_modal_window(WINDOW *parent, int height, int width, int y, int x)
     dialog_state.all_windows = p;
 #ifdef HAVE_COLOR
     if (dialog_state.use_shadow) {
-	draw_childs_shadow(p->shadow = parent, win);
+	p->shadow = parent;
+	draw_childs_shadow(p);
     }
 #endif
 
@@ -1500,19 +2025,12 @@ dlg_new_modal_window(WINDOW *parent, int height, int width, int y, int x)
 void
 dlg_move_window(WINDOW *win, int height, int width, int y, int x)
 {
-    DIALOG_WINDOWS *p, *q;
+    DIALOG_WINDOWS *p;
 
     if (win != 0) {
 	dlg_ctl_size(height, width);
 
-	for (p = dialog_state.all_windows; p != 0; p = q) {
-	    q = p->next;
-	    if (p->normal == win) {
-		break;
-	    }
-	}
-
-	if (p != 0) {
+	if ((p = find_window(win)) != 0) {
 	    (void) wresize(win, height, width);
 	    (void) mvwin(win, y, x);
 #ifdef HAVE_COLOR
@@ -1527,8 +2045,7 @@ dlg_move_window(WINDOW *win, int height, int width, int y, int x)
 	    (void) refresh();
 
 #ifdef HAVE_COLOR
-	    if (p->shadow)
-		draw_childs_shadow(p->shadow, win);
+	    draw_childs_shadow(p);
 #endif
 	}
     }
@@ -1545,6 +2062,7 @@ dlg_sub_window(WINDOW *parent, int height, int width, int y, int x)
 		    y, x, height, width);
     }
 
+    add_subwindow(parent, win);
     (void) keypad(win, TRUE);
     return win;
 }
@@ -1592,7 +2110,7 @@ dlg_default_listitem(DIALOG_LISTITEM * items)
  * Draw the string for item_help
  */
 void
-dlg_item_help(char *txt)
+dlg_item_help(const char *txt)
 {
     if (USE_ITEM_HELP(txt)) {
 	chtype attr = A_NORMAL;
@@ -1606,6 +2124,7 @@ dlg_item_help(char *txt)
 	if (itemhelp_attr & A_COLOR) {
 	    /* fill the remainder of the line with the window's attributes */
 	    getyx(stdscr, y, x);
+	    (void) y;
 	    while (x < COLS) {
 		(void) addch(' ');
 		++x;
@@ -1669,7 +2188,7 @@ dlg_trim_string(char *s)
     char *base = s;
     char *p1;
     char *p = s;
-    int has_newlines = (strstr(s, "\\n") != 0);
+    int has_newlines = !dialog_vars.no_nl_expand && (strstr(s, "\\n") != 0);
 
     while (*p != '\0') {
 	if (*p == TAB && !dialog_vars.nocollapse)
@@ -1775,7 +2294,7 @@ dlg_clr_result(void)
 char *
 dlg_set_result(const char *string)
 {
-    unsigned need = string ? strlen(string) + 1 : 0;
+    unsigned need = string ? (unsigned) strlen(string) + 1 : 0;
 
     /* inputstr.c needs a fixed buffer */
     if (need < MAX_LEN)
@@ -1808,9 +2327,9 @@ void
 dlg_add_result(const char *string)
 {
     unsigned have = (dialog_vars.input_result
-		     ? strlen(dialog_vars.input_result)
+		     ? (unsigned) strlen(dialog_vars.input_result)
 		     : 0);
-    unsigned want = strlen(string) + 1 + have;
+    unsigned want = (unsigned) strlen(string) + 1 + have;
 
     if ((want >= MAX_LEN)
 	|| (dialog_vars.input_length != 0)
@@ -1819,14 +2338,14 @@ dlg_add_result(const char *string)
 	if (dialog_vars.input_length == 0
 	    || dialog_vars.input_result == 0) {
 
-	    char *save = dialog_vars.input_result;
+	    char *save_result = dialog_vars.input_result;
 
 	    dialog_vars.input_length = want * 2;
 	    dialog_vars.input_result = dlg_malloc(char, dialog_vars.input_length);
 	    assert_ptr(dialog_vars.input_result, "dlg_add_result malloc");
-	    dialog_vars.input_result[0] = 0;
-	    if (save != 0)
-		strcpy(dialog_vars.input_result, save);
+	    dialog_vars.input_result[0] = '\0';
+	    if (save_result != 0)
+		strcpy(dialog_vars.input_result, save_result);
 	} else if (want >= dialog_vars.input_length) {
 	    dialog_vars.input_length = want * 2;
 	    dialog_vars.input_result = dlg_realloc(char,
@@ -1838,8 +2357,44 @@ dlg_add_result(const char *string)
     strcat(dialog_vars.input_result, string);
 }
 
-#define FIX_SINGLE "\n\\'"
-#define FIX_DOUBLE "\n\\\"[]{}?*;`~#$^&()|<>\t "
+/*
+ * These are characters that (aside from the quote-delimiter) will have to
+ * be escaped in a single- or double-quoted string.
+ */
+#define FIX_SINGLE "\n\\"
+#define FIX_DOUBLE FIX_SINGLE "[]{}?*;`~#$^&()|<>"
+
+/*
+ * Returns the quote-delimiter.
+ */
+static const char *
+quote_delimiter(void)
+{
+    return dialog_vars.single_quoted ? "'" : "\"";
+}
+
+/*
+ * Returns true if we should quote the given string.
+ */
+static bool
+must_quote(char *string)
+{
+    bool code = FALSE;
+
+    if (*string != '\0') {
+	size_t len = strlen(string);
+	if (strcspn(string, quote_delimiter()) != len)
+	    code = TRUE;
+	else if (strcspn(string, "\n\t ") != len)
+	    code = TRUE;
+	else
+	    code = (strcspn(string, FIX_DOUBLE) != len);
+    } else {
+	code = TRUE;
+    }
+
+    return code;
+}
 
 /*
  * Add a quoted string to the result buffer.
@@ -1848,28 +2403,89 @@ void
 dlg_add_quoted(char *string)
 {
     char temp[2];
-    const char *my_quote = dialog_vars.single_quoted ? "'" : "\"";
+    const char *my_quote = quote_delimiter();
     const char *must_fix = (dialog_vars.single_quoted
 			    ? FIX_SINGLE
 			    : FIX_DOUBLE);
 
-    if (dialog_vars.single_quoted
-	&& strlen(string) != 0
-	&& strcspn(string, FIX_DOUBLE FIX_SINGLE) == strlen(string)) {
-	dlg_add_result(string);
-    } else {
+    if (dialog_vars.quoted || must_quote(string)) {
 	temp[1] = '\0';
 	dlg_add_result(my_quote);
 	while (*string != '\0') {
 	    temp[0] = *string++;
-	    if (strchr(must_fix, *temp) != 0)
+	    if (strchr(my_quote, *temp) || strchr(must_fix, *temp))
 		dlg_add_result("\\");
 	    dlg_add_result(temp);
 	}
 	dlg_add_result(my_quote);
+    } else {
+	dlg_add_result(string);
     }
 }
 
+/*
+ * When adding a result, make that depend on whether "--quoted" is used.
+ */
+void
+dlg_add_string(char *string)
+{
+    if (dialog_vars.quoted) {
+	dlg_add_quoted(string);
+    } else {
+	dlg_add_result(string);
+    }
+}
+
+bool
+dlg_need_separator(void)
+{
+    bool result = FALSE;
+
+    if (dialog_vars.output_separator) {
+	result = TRUE;
+    } else if (dialog_vars.input_result && *(dialog_vars.input_result)) {
+	result = TRUE;
+    }
+    return result;
+}
+
+void
+dlg_add_separator(void)
+{
+    const char *separator = (dialog_vars.separate_output) ? "\n" : " ";
+
+    if (dialog_vars.output_separator)
+	separator = dialog_vars.output_separator;
+
+    dlg_add_result(separator);
+}
+
+/*
+ * Some widgets support only one value of a given variable - save/restore the
+ * global dialog_vars so we can override it consistently.
+ */
+void
+dlg_save_vars(DIALOG_VARS * vars)
+{
+    *vars = dialog_vars;
+}
+
+/*
+ * Most of the data in DIALOG_VARS is normally set by command-line options.
+ * The input_result member is an exception; it is normally set by the dialog
+ * library to return result values.
+ */
+void
+dlg_restore_vars(DIALOG_VARS * vars)
+{
+    char *save_result = dialog_vars.input_result;
+    unsigned save_length = dialog_vars.input_length;
+
+    dialog_vars = *vars;
+    dialog_vars.input_result = save_result;
+    dialog_vars.input_length = save_length;
+}
+
 /*
  * Called each time a widget is invoked which may do output, increment a count.
  */
@@ -1884,14 +2500,14 @@ dlg_does_output(void)
  */
 #if !(defined(HAVE_GETBEGX) && defined(HAVE_GETBEGY))
 int
-getbegx(WINDOW *win)
+dlg_getbegx(WINDOW *win)
 {
     int y, x;
     getbegyx(win, y, x);
     return x;
 }
 int
-getbegy(WINDOW *win)
+dlg_getbegy(WINDOW *win)
 {
     int y, x;
     getbegyx(win, y, x);
@@ -1901,14 +2517,14 @@ getbegy(WINDOW *win)
 
 #if !(defined(HAVE_GETCURX) && defined(HAVE_GETCURY))
 int
-getcurx(WINDOW *win)
+dlg_getcurx(WINDOW *win)
 {
     int y, x;
     getyx(win, y, x);
     return x;
 }
 int
-getcury(WINDOW *win)
+dlg_getcury(WINDOW *win)
 {
     int y, x;
     getyx(win, y, x);
@@ -1918,14 +2534,14 @@ getcury(WINDOW *win)
 
 #if !(defined(HAVE_GETMAXX) && defined(HAVE_GETMAXY))
 int
-getmaxx(WINDOW *win)
+dlg_getmaxx(WINDOW *win)
 {
     int y, x;
     getmaxyx(win, y, x);
     return x;
 }
 int
-getmaxy(WINDOW *win)
+dlg_getmaxy(WINDOW *win)
 {
     int y, x;
     getmaxyx(win, y, x);
@@ -1935,17 +2551,35 @@ getmaxy(WINDOW *win)
 
 #if !(defined(HAVE_GETPARX) && defined(HAVE_GETPARY))
 int
-getparx(WINDOW *win)
+dlg_getparx(WINDOW *win)
 {
     int y, x;
     getparyx(win, y, x);
     return x;
 }
 int
-getpary(WINDOW *win)
+dlg_getpary(WINDOW *win)
 {
     int y, x;
     getparyx(win, y, x);
     return y;
 }
 #endif
+
+#ifdef NEED_WGETPARENT
+WINDOW *
+dlg_wgetparent(WINDOW *win)
+{
+#undef wgetparent
+    WINDOW *result = 0;
+    DIALOG_WINDOWS *p;
+
+    for (p = dialog_state.all_subwindows; p != 0; p = p->next) {
+	if (p->shadow == win) {
+	    result = p->normal;
+	    break;
+	}
+    }
+    return result;
+}
+#endif
